# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Sets up and handles the flow UI context menus"""


import os.path
import uuid
import logging
from ui.qt import QMenu, QApplication
from flowui.cellelement import CellElement
from flowui.items import IfCell
from flowui.scopeitems import ScopeCellElement
from flowui.groupitems import OpenedGroupBegin, CollapsedGroup, EmptyGroup
from flowui.docitems import IndependentDocCell, LeadingDocCell, AboveDocCell
from flowui.cml import CMLVersion, CMLsw, CMLcc, CMLrt, CMLgb, CMLge, CMLdoc
from utils.pixmapcache import getIcon
from utils.diskvaluesrelay import addCollapsedGroup, removeCollapsedGroup
from utils.settings import Settings
from utils.globals import GlobalData
from utils.misc import preResolveLinkPath, getDefaultFileDoc
from .flowuireplacetextdlg import ReplaceTextDialog
from .customcolordlg import CustomColorsDialog
from .flowuidoceditdlg import DocLinkAnchorDialog


class CFSceneContextMenuMixin:

    """Encapsulates the context menu handling"""

    def __init__(self):
        self.menu = None
        self.individualMenus = {}

        # Scene menu preparation
        self.sceneMenu = QMenu()
        self.sceneMenu.addAction(getIcon('filesvg.png'), 'Save as SVG...',
                                 self.parent().onSaveAsSVG)
        self.sceneMenu.addAction(getIcon('filepdf.png'), 'Save as PDF...',
                                 self.parent().onSaveAsPDF)
        self.sceneMenu.addAction(getIcon('filepixmap.png'), 'Save as PNG...',
                                 self.parent().onSaveAsPNG)
        self.sceneMenu.addSeparator()
        self.sceneMenu.addAction(getIcon('copymenu.png'),
                                 'Copy image to clipboard',
                                 self.parent().copyToClipboard)

        # Common menu for all the individually selected items
        self.commonMenu = QMenu()
        self.__ccSubmenuAction = self.commonMenu.addMenu(
            self.__initCustomColorsContextMenu())
        self.__rtSubmenuAction = self.commonMenu.addMenu(
            self.__initReplaceTextContextMenu())
        self.__docSubmenuAction = self.commonMenu.addMenu(
            self.__initDocContextMenu())
        self.__groupAction = self.commonMenu.addAction(
            getIcon("cfgroup.png"), "Group...", self.onGroup)

        #self.commonMenu.addSeparator()
        #self.__cutAction = self.commonMenu.addAction(
        #    getIcon("cutmenu.png"), "Cut (specific for graphics pane)",
        #    self.onCut)
        #self.__copyAction = self.commonMenu.addAction(
        #    getIcon("copymenu.png"), "Copy (specific for graphics pane)",
        #    self.onCopy)
        #self.commonMenu.addSeparator()
        #self.commonMenu.addAction(
        #    getIcon("trash.png"), "Delete", self.onDelete)

        # Individual items specific menu: begin
        ifContextMenu = QMenu()
        ifContextMenu.addAction(
            getIcon("switchbranches.png"), "Switch branch layout",
            self.onSwitchIfBranch)

        self.individualMenus[IfCell] = ifContextMenu
        self.individualMenus[OpenedGroupBegin] = self.__initOpenGroupContextMenu()
        self.individualMenus[CollapsedGroup] = self.__initCloseGroupContextMenu()
        self.individualMenus[EmptyGroup] = self.__initEmptyGroupContextMenu()

        # Individual items specific menu: end

        # Menu for a group of selected items
        self.groupMenu = QMenu()

    def __initOpenGroupContextMenu(self):
        """Creates the open group context menu"""
        ogMenu = QMenu()
        ogMenu.addAction(getIcon("collapse.png"), "Collapse", self.onGroupCollapse)
        ogMenu.addAction(getIcon("replacetitle.png"), "Edit title...", self.onGroupEditTitle)
        ogMenu.addAction(getIcon("ungroup.png"), "Ungroup", self.onGroupUngroup)
        return ogMenu

    def __initCloseGroupContextMenu(self):
        """Creates the closed group context menu"""
        cgMenu = QMenu()
        cgMenu.addAction(getIcon("expand.png"), "Expand", self.onGroupExpand)
        cgMenu.addAction(getIcon("replacetitle.png"), "Edit title...", self.onGroupEditTitle)
        cgMenu.addAction(getIcon("ungroup.png"), "Ungroup", self.onGroupUngroup)
        return cgMenu

    def __initEmptyGroupContextMenu(self):
        """Creates the empty group context menu"""
        egMenu = QMenu()
        egMenu.addAction(getIcon("replacetitle.png"), "Edit title...", self.onGroupEditTitle)
        egMenu.addAction(getIcon("ungroup.png"), "Ungroup", self.onGroupUngroup)
        return egMenu

    def __initCustomColorsContextMenu(self):
        """Create the custom colors submenu"""
        self.__customColorsSubmenu = QMenu('Custom colors')
        self.__customColorsSubmenu.setIcon(getIcon('customcolorsmenu.png'))
        self.__ccAction = self.__customColorsSubmenu.addAction(
            getIcon("customcolors.png"), "Custom colors...",
            self.onCustomColors)
        self.__customColorsSubmenu.addSeparator()
        self.__removeCCAction = self.__customColorsSubmenu.addAction(
            getIcon('trash.png'), 'Remove custom colors',
            self.onRemoveCustomColors)
        return self.__customColorsSubmenu

    def __initReplaceTextContextMenu(self):
        """Create the Replace text submenu"""
        self.__replaceTextSubmenu = QMenu('Replace text')
        self.__replaceTextSubmenu.setIcon(getIcon('replacetextmenu.png'))
        self.__rtAction = self.__replaceTextSubmenu.addAction(
            getIcon("replacetitle.png"), "Replace text...",
            self.onReplaceText)
        self.__replaceTextSubmenu.addSeparator()
        self.__removeRTAction = self.__replaceTextSubmenu.addAction(
            getIcon('trash.png'), 'Remove replacement text',
            self.onRemoveReplacementText)
        return self.__replaceTextSubmenu

    def __initDocContextMenu(self):
        """Create the Documentation submenu"""
        self.__docSubmenu = QMenu('Documentation')
        self.__docSubmenu.setIcon(getIcon('markdown.png'))
        self.__editDocAction = self.__docSubmenu.addAction(
            getIcon('replacetitle.png'), 'Add/edit doc link/anchor...',
            self.onEditDoc)
        self.__autoDocActon = self.__docSubmenu.addAction(
            getIcon('createdoc.png'),
            'Create doc file, add link and open for editing',
            self.onAutoAddDoc)
        self.__docSubmenu.addSeparator()
        self.__removeDocAction = self.__docSubmenu.addAction(
            getIcon('trash.png'), 'Remove doc link/anchor',
            self.onRemoveDoc)
        return self.__docSubmenu

    def onContextMenu(self, event):
        """Triggered when a context menu should be shown"""
        selectedItems = self.selectedItems()
        selectionCount = len(selectedItems)
        if selectionCount == 0:
            self.sceneMenu.popup(event.screenPos())
            return

        if selectionCount == 1:
            self.__buildIndividualMenu(selectedItems[0])
        else:
            self.__buildGroupMenu(selectedItems)
        self.menu.popup(event.screenPos())

    def __destroyDynamicMenu(self):
        """Properly cleans up the menu memory"""
        if self.menu is not None:
            self.menu.deleteLater()
            self.menu = None

    def __buildIndividualMenu(self, item):
        """Builds a context menu for the given item"""
        self.__destroyDynamicMenu()
        self.menu = QMenu()
        if type(item) in self.individualMenus:
            individualPart = self.individualMenus[type(item)]
            self.menu.addActions(individualPart.actions())
            self.menu.addSeparator()
        self.menu.addActions(self.commonMenu.actions())

        # Note: if certain items need to be disabled then it should be done
        #       here
        self.__disableMenuItems()

    def __buildGroupMenu(self, items):
        """Builds a context menu for the group of items"""
        self.__destroyDynamicMenu()
        self.menu = QMenu()
        if type(items[0]) in self.individualMenus:
            if self.areSelectedOfTypes([[items[0].kind, items[0].subKind]]):
                individualPart = self.individualMenus[type(items[0])]
                self.menu.addActions(individualPart.actions())
                self.menu.addSeparator()
        self.menu.addActions(self.commonMenu.actions())
        if not self.groupMenu.isEmpty():
            self.menu.addSeparator()
            self.menu.addActions(self.groupMenu.actions())

        # Note: if certain items need to be disabled then it should be done
        #       here
        self.__disableMenuItems()

    def __disableMenuItems(self):
        """Disables the common menu items as needed"""
        totalComments = self.countComments()
        hasComment = totalComments > 0
        hasDocstring = self.isDocstringInSelection()
        hasMinimizedExcepts = self.isInSelected([(CellElement.EXCEPT_MINIMIZED,
                                                  None)])
        # Doc links are considered comments as well
        totalDocLinks = self.countInSelected([(CellElement.INDEPENDENT_DOC, None),
                                              (CellElement.LEADING_DOC, None),
                                              (CellElement.ABOVE_DOC, None)])
        totalNonDocComments = totalComments - totalDocLinks

        totalGroups = sum(self.countGroups())
        count = len(self.selectedItems())
        totalCCGroups = sum(self.countGroupsWithCustomColors())
        totalCCDocs = self.countDocWithCustomColors()
        minimizedCount = self.countMinimizedItems()

        self.__ccAction.setEnabled(totalNonDocComments == 0 and
                                   not hasMinimizedExcepts)
        self.__removeCCAction.setEnabled(
            self.countItemsWithCML(CMLcc) + totalCCGroups + totalCCDocs == count)
        self.__customColorsSubmenu.setEnabled(self.__ccAction.isEnabled() or
                                              self.__removeCCAction.isEnabled())

        self.__rtAction.setEnabled(not hasComment and
                                   not hasDocstring and
                                   not hasMinimizedExcepts and
                                   totalDocLinks == 0 and
                                   totalGroups == 0)
        self.__removeRTAction.setEnabled(
            self.countItemsWithCML(CMLrt) == count)
        self.__replaceTextSubmenu.setEnabled(self.__rtAction.isEnabled() or
                                             self.__removeRTAction.isEnabled())

        self.__groupAction.setEnabled(self.__canBeGrouped())

        itemsWithDocCML = self.countItemsWithCML(CMLdoc)
        self.__removeDocAction.setEnabled(
            totalDocLinks + itemsWithDocCML == count)
        if count != 1 or totalNonDocComments != 0 or hasDocstring or \
                totalGroups != 0 or minimizedCount > 0:
            self.__editDocAction.setEnabled(False)
            self.__autoDocActon.setEnabled(False)
        else:
            self.__editDocAction.setEnabled(True)
            fileName = None
            editor = self.selectedItems()[0].getEditor()
            if editor:
                fileName = editor._parent.getFileName()
            self.__autoDocActon.setEnabled(
                fileName is not None and totalDocLinks + itemsWithDocCML == 0)
        self.__docSubmenu.setEnabled(self.__removeDocAction.isEnabled() or
                                     self.__editDocAction.isEnabled() or
                                     self.__autoDocActon.isEnabled())

        #self.__cutAction.setEnabled(count == 1)
        #self.__copyAction.setEnabled(count == 1)

    def __actionPrerequisites(self):
        """True if an editor related action can be done"""
        selectedItems = self.selectedItems()
        if not selectedItems:
            return False
        editor = selectedItems[0].getEditor()
        if editor is None:
            return False
        return True

    def onSwitchIfBranch(self):
        """If primitive should switch the branches"""
        if not self.__actionPrerequisites():
            return

        # Memorize the current selection
        selection = self.serializeSelection()

        # The selected items need to be sorted in the reverse line no oreder
        editor = self.selectedItems()[0].getEditor()
        with editor:
            for item in self.sortSelectedReverse():
                if item.kind == CellElement.IF:
                    cmlComment = CMLVersion.find(item.ref.leadingCMLComments,
                                                 CMLsw)
                    if cmlComment is None:
                        # Did not exist, so needs to be generated
                        line = CMLsw.generate(item.ref.body.beginPos)
                        lineNo = item.getFirstLine()
                        editor.insertLines(line, lineNo)
                    else:
                        # Existed, so it just needs to be deleted
                        cmlComment.removeFromText(editor)
        QApplication.processEvents()
        self.parent().redrawNow()
        self.restoreSelectionByID(selection)

    def onCustomColors(self):
        """Custom background and foreground colors"""
        if not self.__actionPrerequisites():
            return

        # Memorize the current selection
        selection = self.serializeSelection()

        bgcolor, fgcolor, bordercolor = self.selectedItems()[0].getColors()
        hasDocstring = self.isDocstringInSelection()
        dlg = CustomColorsDialog(bgcolor, fgcolor,
                                 None if hasDocstring else bordercolor,
                                 self.parent())
        if dlg.exec_():
            bgcolor = dlg.backgroundColor()
            fgcolor = dlg.foregroundColor()
            bordercolor = dlg.borderColor()

            editor = self.selectedItems()[0].getEditor()
            with editor:
                # Add colors is done via delete/insert for the Doc and group
                # items. So it is safer to do first because the cc comment may be
                # in a set of selected which is inserted before the doc cml and
                # thus breaks the line numbering
                for item in self.selectedItems():
                    if item.isCMLDoc():
                        # The doc always exists so just add/change the colors
                        item.cmlRef.updateCustomColors(editor, bgcolor,
                                                       fgcolor, bordercolor)
                        continue
                    if item.isGroupItem():
                        # The group always exists so just add/change the colors
                        item.groupBeginCMLRef.updateCustomColors(editor,
                                                                 bgcolor,
                                                                 fgcolor,
                                                                 bordercolor)

                for item in self.sortSelectedReverse():
                    if item.isCMLDoc() or item.isGroupItem():
                        continue
                    if item.isDocstring():
                        cmlComment = CMLVersion.find(
                            item.ref.docstring.leadingCMLComments,
                            CMLcc)
                    else:
                        cmlComment = CMLVersion.find(
                            item.ref.leadingCMLComments, CMLcc)
                    if cmlComment is not None:
                        # Existed, so remove the old one first
                        lineNo = cmlComment.ref.beginLine
                        cmlComment.removeFromText(editor)
                    else:
                        lineNo = item.getFirstLine()

                    pos = item.ref.body.beginPos
                    if item.isDocstring():
                        pos = item.ref.docstring.beginPos
                    line = CMLcc.generate(bgcolor, fgcolor, bordercolor, pos)
                    editor.insertLines(line, lineNo)
            QApplication.processEvents()
            self.parent().redrawNow()
            self.restoreSelectionByID(selection)

    def onReplaceText(self):
        """Replace the code with a title"""
        if not self.__actionPrerequisites():
            return

        # Memorize the current selection
        selection = self.serializeSelection()

        dlg = ReplaceTextDialog('Replace text', 'Item caption:', self.parent())

        # If it was one item selection and there was a previous text then
        # set it for editing
        previousText = None
        if len(self.selectedItems()) == 1:
            previousText = ''
            cmlComment = CMLVersion.find(
                self.selectedItems()[0].ref.leadingCMLComments, CMLrt)
            if cmlComment is not None:
                previousText = cmlComment.getText()
                dlg.setText(previousText)

        if dlg.exec_():
            replacementText = dlg.text()
            if previousText == replacementText:
                return

            editor = self.selectedItems()[0].getEditor()
            with editor:
                for item in self.sortSelectedReverse():
                    cmlComment = CMLVersion.find(
                        item.ref.leadingCMLComments, CMLrt)
                    if cmlComment is not None:
                        # Existed, so remove the old one first
                        lineNo = cmlComment.ref.beginLine
                        cmlComment.removeFromText(editor)
                    else:
                        lineNo = item.getFirstLine()

                    line = CMLrt.generate(replacementText,
                                          item.ref.body.beginPos)
                    editor.insertLines(line, lineNo)
            QApplication.processEvents()
            self.parent().redrawNow()
            self.restoreSelectionByID(selection)

    def onGroupCollapse(self):
        """Collapses the selected group"""
        if not self.__actionPrerequisites():
            return

        # The selected items need to be sorted in the reverse line no oreder
        editor = self.selectedItems()[0].getEditor()
        with editor:
            for item in self.sortSelectedReverse():
                if item.kind == CellElement.OPENED_GROUP_BEGIN:
                    fileName = editor._parent.getFileName()
                    if not fileName:
                        fileName = editor._parent.getShortName()
                    addCollapsedGroup(fileName, item.getGroupId())

        QApplication.processEvents()
        self.parent().redrawNow()

    def onGroupExpand(self):
        """Expands the selected group"""
        if not self.__actionPrerequisites():
            return

        # The selected items need to be sorted in the reverse line no oreder
        editor = self.selectedItems()[0].getEditor()
        with editor:
            for item in self.sortSelectedReverse():
                if item.kind == CellElement.COLLAPSED_GROUP:
                    fileName = editor._parent.getFileName()
                    if not fileName:
                        fileName = editor._parent.getShortName()
                    removeCollapsedGroup(fileName, item.getGroupId())

        QApplication.processEvents()
        self.parent().redrawNow()

    def onGroupEditTitle(self):
        """Edit (or view) the group title"""
        if not self.__actionPrerequisites():
            return

        # Memorize the current selection
        selection = self.serializeSelection()

        dlg = ReplaceTextDialog('Group title', 'Group title:', self.parent())

        # If it was one item selection and there was a previous text then
        # set it for editing
        if len(self.selectedItems()) == 1:
            title = self.selectedItems()[0].getTitle()
            if title:
                dlg.setText(title)

        if dlg.exec_():
            newTitle = dlg.text()
            editor = self.selectedItems()[0].getEditor()
            with editor:
                for item in self.sortSelectedReverse():
                    item.groupBeginCMLRef.updateTitle(editor, newTitle)
            QApplication.processEvents()
            self.parent().redrawNow()
            self.restoreSelectionByID(selection)

    def onGroupUngroup(self):
        """Ungroups the items"""
        if not self.__actionPrerequisites():
            return

        # Memorize the current selection
        selection = self.serializeSelection()

        # The selected items need to be sorted in the reverse line no oreder
        editor = self.selectedItems()[0].getEditor()
        with editor:
            for item in self.sortSelectedReverse():
                item.groupEndCMLRef.removeFromText(editor)
                item.groupBeginCMLRef.removeFromText(editor)
        QApplication.processEvents()
        self.parent().redrawNow()
        self.restoreSelectionByTooltip(selection)

    def onDelete(self):
        """Delete the item"""
        print("Delete")

    def onGroup(self):
        """Groups items into a single one"""
        dlg = ReplaceTextDialog('Group title', 'Group title:', self.parent())

        if dlg.exec_():
            groupTitle = dlg.text()
            selected = self.__extendSelectionForGrouping()
            selected = self.sortSelected(selected)
            editor = selected[0].getEditor()

            firstLine, lastLine, pos = self.__getLineRange(selected)

            groupid = self.parent().generateNewGroupId()
            beginComment = CMLgb.generate(groupid, groupTitle,
                                          None, None, None, pos)
            endComment = CMLge.generate(groupid, pos)

            with editor:
                editor.insertLines(endComment, lastLine + 1)
                editor.insertLines(beginComment, firstLine)

            # Redraw the group collapsed
            fileName = editor._parent.getFileName()
            if not fileName:
                fileName = editor._parent.getShortName()
            addCollapsedGroup(fileName, groupid)

            QApplication.processEvents()
            self.parent().redrawNow()

    def onCopy(self):
        """Copying..."""
        selectedItems = self.selectedItems()
        if selectedItems:
            if len(selectedItems) > 1:
                print('Copying multiple items has not been implemented yet')
                return
            selectedItems[0].copyToClipboard()

    def onCut(self):
        """Cutting..."""
        print("Cut")

    def onRemoveCustomColors(self):
        """Removing the previously set custom colors"""
        if not self.__actionPrerequisites():
            return

        # Memorize the current selection
        selection = self.serializeSelection()

        editor = self.selectedItems()[0].getEditor()
        with editor:
            # Remove colors is done via delete/insert for the Doc and group
            # items. So it is safer to do first because the cc comment may be
            # in a set of selected which is inserted before the doc cml and
            # thus breaks the line numbering
            for item in self.selectedItems():
                # The doc always exists
                if item.isCMLDoc():
                    item.cmlRef.removeCustomColors(editor)
                    continue
                # The group always exists
                if item.isGroupItem():
                    item.groupBeginCMLRef.removeCustomColors(editor)

            # Now handle the rest of items
            for item in self.sortSelectedReverse():
                if item.isCMLDoc() or item.isGroupItem():
                    continue
                if item.isDocstring():
                    cmlComment = CMLVersion.find(
                        item.ref.docstring.leadingCMLComments, CMLcc)
                else:
                    cmlComment = CMLVersion.find(
                        item.ref.leadingCMLComments, CMLcc)
                if cmlComment is not None:
                    cmlComment.removeFromText(editor)
        QApplication.processEvents()
        self.parent().redrawNow()
        self.restoreSelectionByID(selection)

    def onRemoveReplacementText(self):
        """Removing replacement text"""
        if self.__actionPrerequisites():
            # Memorize the current selection
            selection = self.serializeSelection()

            editor = self.selectedItems()[0].getEditor()
            with editor:
                for item in self.sortSelectedReverse():
                    cmlComment = CMLVersion.find(item.ref.leadingCMLComments,
                                                 CMLrt)
                    if cmlComment is not None:
                        cmlComment.removeFromText(editor)
            QApplication.processEvents()
            self.parent().redrawNow()
            self.restoreSelectionByID(selection)

    def areSelectedOfTypes(self, matchList):
        """Checks if the selected items belong to the match"""
        # match is a list of pairs [kind, subKind]
        #   None would mean 'match any'
        selectedItems = self.selectedItems()
        if selectedItems:
            for selectedItem in selectedItems:
                for kind, subKind in matchList:
                    match = True
                    if kind is not None:
                        if kind != selectedItem.kind:
                            match = False
                    if subKind is not None:
                        if subKind != selectedItem.subKind:
                            match = False
                    if match:
                        break
                else:
                    return False
            return True
        return False

    def __createDocFile(self, link, fromFile):
        """Creates the doc file if needed"""
        fName, _, errMsg = preResolveLinkPath(link, fromFile, True)
        if errMsg:
            logging.error(errMsg)
            return None

        if os.path.exists(fName):
            return fName

        try:
            os.makedirs(os.path.dirname(fName), exist_ok=True)
            with open(fName, 'w') as f:
                pass
        except Exception as exc:
            logging.error('Error creating the documentation file ' +
                          fName + ': ' + str(exc))
            return None
        return fName

    def onEditDoc(self):
        """Editing the CML doc comment"""
        if not self.__actionPrerequisites():
            return

        selectedItem = self.selectedItems()[0]  # Exactly one is selected
        editor = selectedItem.getEditor()
        fileName = editor._parent.getFileName()
        if not fileName:
            fileName = editor._parent.getShortName()

        # It could be a CML doc or an item which has a CML doc
        if selectedItem.isComment():
            cmlRef = selectedItem.cmlRef
        else:
            # If not found then it means the doc link needs to be created
            cmlRef = self.__findCMLinItem(selectedItem, CMLdoc)

        dlg = DocLinkAnchorDialog('Add' if cmlRef is None else 'Edit',
                                  cmlRef, fileName, self.parent())
        if dlg.exec_():
            link = dlg.linkEdit.text().strip()
            anchor = dlg.anchorEdit.text().strip()
            title = dlg.title()
            needToCreate = dlg.needToCreate()

            # First create a file if asked
            if needToCreate:
                docFileName = self.__createDocFile(link, fileName)
                if not docFileName:
                    return

            selection = self.serializeSelection()
            with editor:
                # Now insert a new cml comment or update existing
                if cmlRef:
                    # It is editing, the comment exists
                    lineNo = cmlRef.ref.beginLine
                    pos = cmlRef.ref.beginPos
                    cmlRef.removeFromText(editor)
                    bgColor = cmlRef.bgColor
                    fgColor = cmlRef.fgColor
                    border = cmlRef.border
                else:
                    # It is a new doc link
                    lineNo = selectedItem.getFirstLine()
                    pos = selectedItem.ref.body.beginPos
                    bgColor = None
                    fgColor = None
                    border = None

                line = CMLdoc.generate(link, anchor, title,
                                       bgColor, fgColor, border, pos)
                editor.insertLines(line, lineNo)

            QApplication.processEvents()
            self.parent().redrawNow()
            self.restoreSelectionByID(selection)

    @staticmethod
    def __getAutoDocFileName(fileName):
        """Forms the auto doc file name"""
        # Markdown is used as a default documentation format
        fBaseName = os.path.basename(fileName)
        if '.' in fBaseName:
            fileExtension = fBaseName.split('.')[-1]
            fBaseName = fBaseName[:-len(fileExtension)] + 'md'
        else:
            fBaseName += '.md'

        project = GlobalData().project
        if project.isProjectFile(fileName):
            projectDir = project.getProjectDir()
            relativePath = fileName[len(projectDir):]
            projectName = project.getProjectName()
            if relativePath.startswith(projectName):
                relativePath = relativePath.replace(projectName, '', 1)
            return os.path.normpath(
                os.path.sep.join([projectDir + 'doc',
                                  os.path.dirname(relativePath),
                                  fBaseName]))
        return os.path.normpath(
            os.path.sep.join([os.path.dirname(fileName),
                              'doc',
                               fBaseName]))

    def onAutoAddDoc(self):
        """Create a doc file, add a link and open for editing"""
        if not self.__actionPrerequisites():
            return

        selectedItem = self.selectedItems()[0]  # Exactly one is selected
        editor = selectedItem.getEditor()
        fileName = editor._parent.getFileName()
        if not fileName:
            logging.error('Save file before invoking auto doc')
            return

        needContent = False
        newAnchor = 'doc' + str(uuid.uuid4().fields[-1])[-6:]

        docFileName = self.__getAutoDocFileName(fileName)
        if not os.path.exists(docFileName):
            # Create file and populate with the default content
            try:
                os.makedirs(os.path.dirname(docFileName), exist_ok=True)
                with open(docFileName, 'w') as f:
                    pass
            except Exception as exc:
                logging.error('Error creating the documentation file ' +
                              docFileName + ': ' + str(exc))
                return
            needContent = True

        project = GlobalData().project
        if project.isProjectFile(docFileName):
            linkFromFile = project.getRelativePath(docFileName)
            if project.isProjectFile(fileName):
                linkFromDoc = project.getRelativePath(fileName)
            else:
                linkFromDoc = os.path.relpath(fileName,
                                              os.path.dirname(docFileName))
        else:
            linkFromFile = os.path.relpath(docFileName,
                                           os.path.dirname(fileName))
            linkFromDoc = os.path.relpath(fileName,
                                          os.path.dirname(docFileName))


        # Insert a doc link
        with editor:
            if self.__isModuleSelected():
                lineNo = 1
                if selectedItem.ref.encodingLine:
                    lineNo += 1
                if selectedItem.ref.bangLine:
                    lineNo += 1
            else:
                lineNo = selectedItem.getFirstLine()
            line = CMLdoc.generate(linkFromFile, newAnchor, 'See documentation',
                                   None, None, None,
                                   selectedItem.ref.body.beginPos)
            editor.insertLines(line, lineNo)

            QApplication.processEvents()
            self.parent().redrawNow()

        # Open the file
        if GlobalData().mainWindow.openFile(docFileName, -1):
            if needContent:
                widget = GlobalData().mainWindow.em.getWidgetForFileName(docFileName)
                editor = widget.getEditor()
                editor.text = getDefaultFileDoc(linkFromDoc, newAnchor)
                editor.document().setModified(False)

    def onRemoveDoc(self):
        """Removing the CML doc comment"""
        if not self.__actionPrerequisites():
            return

        editor = self.selectedItems()[0].getEditor()
        with editor:
            for item in self.sortSelectedReverse():
                cmlComment = CMLVersion.find(item.ref.leadingCMLComments,
                                             CMLdoc)
                if cmlComment is not None:
                    cmlComment.removeFromText(editor)
        QApplication.processEvents()
        self.parent().redrawNow()

    def countInSelected(self, matchList):
        """Counts the number of matching items in selection"""
        # match is a list of pairs [kind, subKind]
        #   None would mean 'match any'
        count = 0
        for selectedItem in self.selectedItems():
            for kind, subKind in matchList:
                match = True
                if kind is not None:
                    if kind != selectedItem.kind:
                        match = False
                if subKind is not None:
                    if subKind != selectedItem.subKind:
                        match = False
                if match:
                    count += 1
        return count

    def isInSelected(self, matchList):
        """Checks if any if the match list items is in the selection"""
        return self.countInSelected(matchList) > 0

    def isDocstringInSelection(self):
        """True if a docstring item in the selection"""
        for item in self.selectedItems():
            if item.isDocstring():
                return True
        return False

    def countComments(self):
        """Count comments in selection"""
        count = 0
        for item in self.selectedItems():
            if item.isComment():
                count += 1
        return count

    def isCommentInSelection(self):
        """True if a comment item in the selection"""
        return self.countComments() > 0

    def countItemsWithCML(self, cmlType):
        """Counts items with have a certain type of a CML comment"""
        count = 0
        for item in self.selectedItems():
            if self.__findCMLinItem(item, cmlType) is not None:
                count += 1
        return count

    def __findCMLinItem(self, item, cmlType):
        """Finds a related CML item"""
        if item.isComment():
            # Doc links are comments so they are skipped here
            return None
        if item.isDocstring():
            # Side comments for docstrings? Nonesense! So they are ignored
            # even if they are collected
            cml = CMLVersion.find(item.ref.docstring.leadingCMLComments, cmlType)
            if cml is not None:
                return cml

        if hasattr(item.ref, 'leadingCMLComments'):
            cml = CMLVersion.find(item.ref.leadingCMLComments, cmlType)
            if cml is not None:
                return cml
        if hasattr(item.ref, 'sideCMLComments'):
            cml = CMLVersion.find(item.ref.sideCMLComments, cmlType)
            if cml is not None:
                return cml
        return None

    def countGroups(self):
        """Counts empty, close and open groups"""
        emptyCount = 0
        closeCount = 0
        openCount = 0
        for item in self.selectedItems():
            if item.kind == CellElement.EMPTY_GROUP:
                emptyCount += 1
            elif item.kind == CellElement.COLLAPSED_GROUP:
                closeCount += 1
            elif item.kind == CellElement.OPENED_GROUP_BEGIN:
                openCount += 1
        return emptyCount, closeCount, openCount

    def countGroupsWithCustomColors(self):
        """Counts the groups with any color defined"""
        emptyCount = 0
        closeCount = 0
        openCount = 0
        for item in self.selectedItems():
            if item.kind in [CellElement.EMPTY_GROUP,
                             CellElement.COLLAPSED_GROUP,
                             CellElement.OPENED_GROUP_BEGIN]:
                if item.groupBeginCMLRef.bgColor is not None or \
                   item.groupBeginCMLRef.fgColor is not None or \
                   item.groupBeginCMLRef.border is not None:
                    if item.kind == CellElement.EMPTY_GROUP:
                        emptyCount += 1
                    elif item.kind == CellElement.COLLAPSED_GROUP:
                        closeCount += 1
                    else:
                        openCount += 1
        return emptyCount, closeCount, openCount

    def countDocWithCustomColors(self):
        count = 0
        for item in self.selectedItems():
            if item.isCMLDoc():
                if item.cmlRef.bgColor is not None or \
                   item.cmlRef.fgColor is not None or \
                   item.cmlRef.border is not None:
                    count += 1
        return count

    def countMinimizedItems(self):
        count = 0
        for item in self.selectedItems():
            if item.isMinimizedItem():
                count += 1
        return count

    def sortSelectedReverse(self):
        """Sorts the selected items in reverse order"""
        result = []
        for item in self.selectedItems():
            itemBegin = item.getAbsPosRange()[0]
            for index in range(len(result)):
                if itemBegin > result[index].getAbsPosRange()[0]:
                    result.insert(index, item)
                    break
            else:
                result.append(item)
        return result

    def sortSelected(self, selected):
        """Sorts the selected items in direct order"""
        result = []
        for item in selected:
            itemBegin = item.getAbsPosRange()[0]
            for index in range(len(result)):
                if itemBegin < result[index].getAbsPosRange()[0]:
                    result.insert(index, item)
                    break
            else:
                result.append(item)
        return result

    def __canBeGrouped(self):
        """True if the selected items can be grouped"""
        # Cannot import it at the top...
        from .flowuiwidget import SMART_ZOOM_ALL, SMART_ZOOM_NO_CONTENT

        if Settings()['smartZoom'] not in [SMART_ZOOM_ALL,
                                           SMART_ZOOM_NO_CONTENT]:
            return False
        if self.__areAllSelectedComments():
            return False
        if self.__areScopeDocstringOrCommentSelected():
            return False
        if self.__isModuleSelected():
            return False

        # Extend the selection with all the selected items comments
        selected = self.__extendSelectionForGrouping()

        if self.__areLoneCommentsSelected(selected):
            return False

        if self.__areIncompleteScopeSelected(selected):
            return False

        scopeCoveredRegions = self.__getSelectedScopeRegions(selected)

        # The __areIfFullySelected() also updates the regions with
        # fully selected if regions
        if not self.__areIfFullySelected(selected, scopeCoveredRegions):
            return False

        selected = self.sortSelected(selected)
        begin = selected[0].getAbsPosRange()[0]
        end = selected[-1].getAbsPosRange()[1]

        if not self.__isSelectionContinuous(selected, scopeCoveredRegions,
                                            begin, end):
            return False

        if self.__moreThanOneIfBranchSelected(selected, scopeCoveredRegions):
            return False
        return True

    def __areAllSelectedComments(self):
        """True if all selected items are comments"""
        for item in self.selectedItems():
            if not item.isComment():
                return False
        return True

    def __areScopeDocstringOrCommentSelected(self):
        for item in self.selectedItems():
            if item.scopedItem():
                if item.subKind in [ScopeCellElement.COMMENT,
                                    ScopeCellElement.DOCSTRING]:
                    return True
        return False

    def __isModuleSelected(self):
        """True if the whole module is selected"""
        for item in self.selectedItems():
            if item.kind == CellElement.FILE_SCOPE:
                return True
        return False

    def __areIncompleteScopeSelected(self, selected):
        """True if an incomplete scope selected"""
        for item in selected:
            if item.kind in [CellElement.FOR_SCOPE,
                             CellElement.WHILE_SCOPE]:
                if item.ref.elsePart:
                    for relatedItem in self.findItemsForRef(item.ref.elsePart):
                        if relatedItem not in selected:
                            return True
            elif item.kind in [CellElement.TRY_SCOPE]:
                # It could be that the exception blocks are hidden, so there
                # will be exactly one more item instead of many and that item
                # will have a ref which matches the try scope.
                exceptPartCount = 0
                for exceptPart in item.ref.exceptParts:
                    for relatedItem in self.findItemsForRef(exceptPart):
                        exceptPartCount += 1
                        if relatedItem not in selected:
                            return True
                if exceptPartCount == 0:
                    # here: no except blocks on the diagram, they are collapsed
                    tryItems = self.findItemsForRef(item.ref)
                    for tryItem in tryItems:
                        if tryItem.kind == CellElement.EXCEPT_MINIMIZED:
                            if not tryItem.isSelected():
                                return True
                            break
                    else:
                        # The minimized except is not selected
                        return True
                if item.ref.elsePart:
                    for relatedItem in self.findItemsForRef(item.ref.elsePart):
                        if relatedItem not in selected:
                            return True
                if item.ref.finallyPart:
                    for relatedItem in self.findItemsForRef(item.ref.finallyPart):
                        if relatedItem not in selected:
                            return True
            elif item.kind in [CellElement.ELSE_SCOPE,
                               CellElement.EXCEPT_SCOPE,
                               CellElement.FINALLY_SCOPE]:
                for relatedItem in self.findItemsForRef(item.leaderRef):
                    if relatedItem not in selected:
                        return True
            elif item.kind == CellElement.EXCEPT_MINIMIZED:
                # here: no except blocks on the diagram, they are collapsed
                tryItems = self.findItemsForRef(item.ref)
                for tryItem in tryItems:
                    if tryItem.kind == CellElement.TRY_SCOPE:
                        if tryItem.subKind == ScopeCellElement.TOP_LEFT:
                            if not tryItem.isSelected():
                                return True
                            break
                else:
                    # The try is not selected
                    return True
        return False

    def __extendSelectionForGrouping(self):
        """Extends the selection with the leading and side comments"""
        boundComments = []
        selected = self.selectedItems()
        for item in selected:
            if not item.isComment() and not self.isOpenGroupItem(item):
                for relatedItem in self.findItemsForRef(item.ref):
                    if relatedItem not in selected:
                        boundComments.append(relatedItem)
        return selected + boundComments

    def __areLoneCommentsSelected(self, selected):
        """True if there are comments selected which have no main item selected"""
        for item in selected:
            if item.isComment():
                if item.kind in [CellElement.SIDE_COMMENT,
                                 CellElement.LEADING_COMMENT,
                                 CellElement.ABOVE_COMMENT]:
                    for relatedItem in self.findItemsForRef(item.ref):
                        if relatedItem not in selected:
                            return True
        return False

    def __getLineRange(self, selected):
        first = selected[0]
        last = selected[-1]

        if first.kind == CellElement.OPENED_GROUP_BEGIN:
            firstLine = first.groupBeginCMLRef.ref.parts[0].beginLine
            pos = first.groupBeginCMLRef.ref.parts[0].beginPos
        else:
            firstLine = first.getLineRange()[0]
            pos = first.ref.beginPos

        if last.scopedItem():
            lastLine = last.ref.endLine
        elif last.kind == CellElement.OPENED_GROUP_BEGIN:
            lastLine = last.groupEndCMLRef.ref.parts[-1].endLine
        else:
            lastLine = last.getLineRange()[1]
        return firstLine, lastLine, pos

    def __getSelectedScopeRegions(self, selected):
        """Provides the regions of the selected scope items"""
        coveredRegions = []
        for item in selected:
            if item.scopedItem():
                if item.subKind in [ScopeCellElement.TOP_LEFT]:
                    if item.ref.leadingComment:
                        coveredRegions.append((item.ref.leadingComment.begin,
                                               item.ref.end))
                    else:
                        coveredRegions.append((item.ref.begin, item.ref.end))
            elif item.kind == CellElement.OPENED_GROUP_BEGIN:
                coveredRegions.append(item.getAbsPosRange())
        return coveredRegions

    def __areIfFullySelected(self, selected, regions):
        """Checks if selected IFs are fully selected"""
        for item in selected:
            if item.kind == CellElement.IF:
                ifBegin = item.ref.begin
                ifEnd = item.ref.end
                for item in self.items():
                    if not isinstance(item, CellElement):
                        continue
                    if item.isProxyItem():
                        continue
                    if item.scopedItem():
                        if item.subKind not in [ScopeCellElement.TOP_LEFT,
                                                ScopeCellElement.DOCSTRING,
                                                ScopeCellElement.COMMENT]:
                            continue
                    if item in selected:
                        continue
                    itemRange = item.getAbsPosRange()
                    if self.isInRegion(itemRange[0], itemRange[1], regions):
                        continue
                    if itemRange[0] > ifBegin and itemRange[0] < ifEnd:
                        return False
                    if itemRange[1] > ifBegin and itemRange[1] < ifEnd:
                        return False
                regions.append([ifBegin, ifEnd])
        return True

    @staticmethod
    def isInRegion(start, finish, regions):
        for region in regions:
            if start >= region[0] and finish <= region[1]:
                return True
        return False

    def __isSelectionContinuous(self, selected, regions, begin, end):
        """Checks if the selection is continuous"""
        for item in self.items():
            if not isinstance(item, CellElement):
                continue
            if item.isProxyItem():
                continue
            if item.scopedItem():
                if item.subKind not in [ScopeCellElement.TOP_LEFT,
                                        ScopeCellElement.DOCSTRING,
                                        ScopeCellElement.COMMENT]:
                    continue
            if item in selected:
                continue
            itemRange = item.getAbsPosRange()
            if self.isInRegion(itemRange[0], itemRange[1], regions):
                continue

            # It is important to keep < and > instead of <= and >=
            # This is because the scopes start with the first statement
            if itemRange[0] > begin and itemRange[0] < end:
                return False
            if itemRange[1] > begin and itemRange[1] < end:
                return False
        return True

    def __moreThanOneIfBranchSelected(self, selected, regions):
        """Checks that the continuous selected items belong to more than one
        not selected IF statements
        """
        ifRef = None
        for item in selected:
            if item.kind != CellElement.IF:
                itemRange = item.getAbsPosRange()
                if item.kind != CellElement.OPENED_GROUP_BEGIN:
                    if self.isInRegion(itemRange[0], itemRange[1], regions):
                        # Here: an item is in a selected scope item, in a selected
                        #       open group or in a fully selected if
                        continue
                # Test if an item belongs to an if statement branch
                if item.kind in [CellElement.OPENED_GROUP_BEGIN,
                                 CellElement.EMPTY_GROUP,
                                 CellElement.COLLAPSED_GROUP]:
                    branchId = item.groupBeginCMLRef.ref.getParentIfID()
                elif item.kind in [CellElement.INDEPENDENT_DOC,
                                   CellElement.LEADING_DOC,
                                   CellElement.ABOVE_DOC]:
                    branchId = item.cmlRef.ref.getParentIfID()
                else:
                    branchId = item.ref.getParentIfID()
                if branchId is not None:
                    if ifRef is None:
                        ifRef = branchId
                    else:
                        if branchId != ifRef:
                            # Selected items belong to more than one branch
                            return True
        return False

    def terminateMenus(self):
        """Called when a tab is closed"""
        self.sceneMenu.deleteLater()
        self.commonMenu.deleteLater()
        self.groupMenu.deleteLater()
        self.__destroyDynamicMenu()

        for menu in self.individualMenus.values():
            menu.deleteLater()




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Wizard plugin interface"""

from .cdmpluginbase import CDMPluginBase


class WizardInterface(CDMPluginBase):

    """Wizard plugin interface"""

    def __init__(self):
        """The plugin class is instantiated with no arguments.

        Instantiating is done regardless wheather a plugin is
        enabled or disabled. So it is recommended to have the
        resource allocation in the activate(...) method and
        deallocation of them in the deactivate(...) method.
        """
        CDMPluginBase.__init__(self)

    # Member functions below could or should be implemented by a plugin.
    # See docstrings for the detailed description.

    @staticmethod
    def isIDEVersionCompatible(ideVersion):
        """Codimension makes this call before activating a plugin.

        The passed ideVersion is a string representing
        the current IDE version.
        True should be returned if the plugin is compatible with the IDE.
        """
        raise Exception("isIDEVersionCompatible() must be overridden")

    def activate(self, ideSettings, ideGlobalData):
        """Can be may overriden to do specific plugin activation handling.

        ideSettings - reference to the IDE Settings singleton
                      see codimension/src/utils/settings.py
        ideGlobalData - reference to the IDE global settings
                        see codimension/src/utils/globals.py

        Note: if overridden then call the base class activate() first.
              Plugin specific activation handling should follow it.
        """
        CDMPluginBase.activate(self, ideSettings, ideGlobalData)

    def deactivate(self):
        """Can be overriden to do specific plugin deactivation handling.

        Note: if overridden then first do the plugin specific deactivation
              handling and then call the base class deactivate()
        """
        CDMPluginBase.deactivate(self)

    def getConfigFunction(self):
        """Provides the config function.

        The plugin can provide a function which will be called when the
        user requests plugin configuring.
        If a plugin does not require any config parameters then None
        should be returned.
        By default no configuring is required.
        """
        return CDMPluginBase.getConfigFunction(self)

    def populateMainMenu(self, parentMenu):
        """Called to populate the main menu.

        The main menu looks as follows:
        Plugins
            - Plugin manager (fixed item)
            - Separator (visible if at lease one plugin provided its
              main menu)
            - <Plugin #1 name> (this is the parentMenu passed)
            ...
        If no items were populated by the plugin then there will be no
        <Plugin #N name> menu item shown.
        Codimension will remove the populated menu when
        a plugin is disabled.
        """
        pass

    def populateFileContextMenu(self, parentMenu):
        """Called to populate the file menu.

        The file context menu shown in the project viewer window will have
        an item with a plugin name and subitems which are populated here.
        If no items were populated then the plugin menu item will not be
        shown.
        When a callback is called the parentMenu data will be set to the
        absolute path of the selected item. So the plugin will be able
        to retrieve the selected path as follows:
        # In this method
        self.fileParentMenu = parentMenu
        ...
        # In a menu item handler
        path = str( self.fileParentMenu.menuAction().data().toString() )

        Codimension will remove the populated menu when
        a plugin is disabled.
        """
        pass

    def populateDirectoryContextMenu(self, parentMenu):
        """Called to populate the directory menu.

        The directory context menu shown in the project viewer window
        will have an item with a plugin name and subitems which are
        populated here.
        If no items were populated then the plugin menu item will not be
        shown.
        When a callback is called the parentMenu data will be set to the
        absolute path of the selected item. So the plugin will be able
        to retrieve the selected path as follows:
        # In this method
        self.dirParentMenu = parentMenu
        ...
        # In a menu item handler
        path = str( self.dirParentMenu.menuAction().data().toString() )

        Codimension will remove the populated menu when
        a plugin is disabled.
        """
        pass

    def populateBufferContextMenu(self, parentMenu):
        """Called to populate the editing buffer menu.

        The buffer context menu shown for the current edited/viewed file
        will have an item with a plugin name and subitems which are
        populated here.
        If no items were populated then the plugin menu item will not be
        shown.
        Codimension will remove the populated menu when
        a plugin is disabled.
        Note: when a buffer context menu is selected by the user it always
              refers to the current widget. To get access to the current
              editing widget the plugin can use:
              self.ide.currentEditorWidget
              The widget could be of different types and some circumstances
              should be considered, e.g.:
              - it could be a new file which has not been saved yet
              - it could be modified
              - it could be that the disk file has already been deleted
              - etc.
              Having the current widget reference the plugin is able to
              retrieve the information it needs."""
        pass



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2020  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


"""Varous specialized labels used in the IDE"""

from utils.fileutils import compactPath
from .qt import Qt, QLabel, QApplication, pyqtSignal


class FitLabel(QLabel):

    """a label that squeezes its contents to fit it's size"""

    def __init__(self, parent=None):
        QLabel.__init__(self, parent)
        self.__text = ''
        self.__customTooltip = None

    def paintEvent(self, event):
        """Called when painting is required"""
        metric = self.fontMetrics()
        if metric.width(self.__text) > self.contentsRect().width():
            QLabel.setText(self,
                           metric.elidedText(self.__text,
                                             Qt.ElideMiddle, self.width()))
            if self.__customTooltip is None:
                QLabel.setToolTip(self, self.__text)
        else:
            QLabel.setText(self, self.__text)
            if self.__customTooltip is None:
                QLabel.setToolTip(self, '')

        QLabel.paintEvent(self, event)

    def setText(self, txt):
        """Set the text to be shown"""
        self.__text = txt
        QLabel.setText(self, txt)

    def setToolTip(self, tooltip):
        """Sets the custom tooltip"""
        self.__customTooltip = tooltip
        if tooltip is None:
            tooltip = ''
        QLabel.setToolTip(self, tooltip)


class FramedFitLabel(FitLabel):

    """Subclassing for CSS styling"""

    def __init__(self, parent=None):
        FitLabel.__init__(self, parent)


class HeaderFitLabel(FitLabel):

    """Subclassing for CSS styling"""

    def __init__(self, parent=None):
        FitLabel.__init__(self, parent)


# Default double click handler is to copy the text to the buffer
class DoubleClickLabel(QLabel):

    """A label with double click event support"""

    doubleClicked = pyqtSignal()

    def __init__(self, text=None, callback=None, parent=None):
        QLabel.__init__(self, parent)
        if text is not None:
            self.setText(text)
        self.__callback = callback

    def mouseDoubleClickEvent(self, event):
        """Handles the mouse double click"""
        if event.button() == Qt.LeftButton:
            if self.__callback is None:
                txt = self.text().strip()
                if txt:
                    QApplication.clipboard().setText(txt)
            elif isinstance(self.__callback, str):
                if self.__callback.lower() == 'signal':
                    self.doubleClicked.emit()
                elif self.__callback.lower() == 'ignore':
                    pass
                else:
                    raise Exception(
                        'Unsupported callback value "' + self.__callback +
                        '". Supported values are: "signal" and "ignore".')
            else:
                self.__callback()
        QLabel.mouseDoubleClickEvent(self, event)


# Supposed to have styled:
# - transparent background
# - border color
# - spacing
# Also implements the default double click copying
class FramedLabel(DoubleClickLabel):

    """A label with a frame and double click for copy content to clipboard"""

    def __init__(self, text=None, callback=None, parent=None):
        DoubleClickLabel.__init__(self, text, callback, parent)


# Supposed to have styled:
# - background color
# - border color
# - spacing
# Also implements the default double click copying
class HeaderLabel(DoubleClickLabel):

    """Subclassed for styling via CSS"""

    def __init__(self, text=None, callback=None, parent=None):
        DoubleClickLabel.__init__(self, text, callback, parent)


class FitPathLabel(DoubleClickLabel):

    """a label showing a file path compacted to fit it's size"""

    def __init__(self, callback=None, parent=None):
        DoubleClickLabel.__init__(self, None, callback, parent)
        self.__path = ''

    def setPath(self, path):
        """Set the path to be shown"""
        self.__path = path
        QLabel.setText(self, path)

    def getPath(self):
        """Provides the stored path"""
        return self.__path

    def paintEvent(self, event):
        """Called when painting is required"""
        sparePixels = 5
        metric = self.fontMetrics()
        requiredWidth = metric.width(self.__path)
        if requiredWidth > self.contentsRect().width() - sparePixels:
            compacted = compactPath(self.__path,
                                    self.contentsRect().width() - sparePixels,
                                    self.length)
            QLabel.setText(self, compacted)
            self.setToolTip(self.__path)
        else:
            QLabel.setText(self, self.__path)
            self.setToolTip('')
        QLabel.paintEvent(self, event)

    def length(self, txt):
        """Length of a text in pixels"""
        return self.fontMetrics().width(txt)



class FramedFitPathLabel(FitPathLabel):

    """a label showing a file path compacted to fit it's size"""

    def __init__(self, callback=None, parent=None):
        FitPathLabel.__init__(self, callback, parent)


class HeaderFitPathLabel(FitPathLabel):

    """a label showing a file path compacted to fit it's size"""

    def __init__(self, callback=None, parent=None):
        FitPathLabel.__init__(self, callback, parent)


# Status bar items

# Supposed to have styled:
# - background color
# - border color
# - spacing
# Also implements the default double click copying
class StatusBarFramedLabel(FramedLabel):

    """Subclassed for styling via CSS"""

    def __init__(self, text=None, callback=None, parent=None):
        FramedLabel.__init__(self, text, callback, parent)



# Supposed to have styled:
# - background: transparent
# - no border
class StatusBarPixmapLabel(DoubleClickLabel):

    """Used for items which display an icon"""

    def __init__(self, callback, parent):
        DoubleClickLabel.__init__(self, None, callback, parent)


# Supposed to have styled:
# - background color
# - border color
# - spacing
class StatusBarPathLabel(FitPathLabel):

    """Subclassed for styling via CSS"""

    def __init__(self, callback, parent):
        FitPathLabel.__init__(self, callback, parent)




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tabs history support implementation"""

import os.path
from .qt import QObject, pyqtSignal
from .mainwindowtabwidgetbase import MainWindowTabWidgetBase


class TabHistoryEntry:

    """Holds a single history entry"""

    def __init__(self):
        self.icon = None
        self.uuid = None
        self.displayName = ""
        self.line = -1
        self.pos = -1
        self.firstVisible = -1
        self.tabType = MainWindowTabWidgetBase.Unknown

    def __eq__(self, other):
        """Compares two entries"""
        return self.uuid == other.uuid and \
               self.line == other.line and \
               self.pos == other.pos


class TabsHistory(QObject):

    """Holds the editors manager history"""

    historyChanged = pyqtSignal()
    limit = 32

    def __init__(self, editorsManager):
        QObject.__init__(self)

        self.__editorsManger = editorsManager
        self.__history = []

        # The index always points to the history position which is currently
        # displayed
        self.__index = -1

        # Sequence of the history entries as they have appeared on the screen
        self.__tabsSequence = []

    def backAvailable(self):
        """True if step back available"""
        return self.__index > 0

    def forwardAvailable(self):
        """True if step forward available"""
        if self.__index == -1:
            return False
        return self.__index < len(self.__history) - 1

    def clear(self):
        """Clears the history"""
        self.__history = []
        self.__index = -1
        self.__tabsSequence = []
        self.historyChanged.emit()

    def addCurrent(self):
        """Adds the current editors manager tab to the history"""
        currentWidget = self.__editorsManger.currentWidget()

        newEntry = TabHistoryEntry()
        newEntry.tabType = currentWidget.getType()
        newEntry.displayName = currentWidget.getShortName()

        newEntry.icon = self.__editorsManger.tabIcon(
            self.__editorsManger.currentIndex())
        newEntry.uuid = currentWidget.getUUID()

        if newEntry.tabType in [MainWindowTabWidgetBase.PlainTextEditor,
                                MainWindowTabWidgetBase.VCSAnnotateViewer]:
            newEntry.line = currentWidget.getLine()
            newEntry.pos = currentWidget.getPos()
            newEntry.firstVisible = currentWidget.getEditor().firstVisibleLine()

        if self.__index != -1:
            if newEntry == self.__history[self.__index]:
                # The new entry is the same as the current - ignore request
                return
            # Cut the tail of the history if needed
            self.__history = self.__history[:self.__index + 1]

        self.__history.append(newEntry)
        self.__enforceLimit()

        self.__index = len(self.__history) - 1
        self.__tabsSequence.append(self.__index)

        self.historyChanged.emit()

    def __enforceLimit(self):
        """Dismisses too old records if required"""
        if len(self.__history) <= self.limit:
            return

        stripCount = len(self.__history) - self.limit
        index = 0
        while index < stripCount:
            # The indexes must be removed from the tabs seq list
            seqIndex = len(self.__tabsSequence) - 1
            while seqIndex >= 0:
                if self.__tabsSequence[seqIndex] == index:
                    del self.__tabsSequence[seqIndex]
                seqIndex -= 1
            index += 1

        # Adjust the indexes in the tabs seq list
        seqIndex = len(self.__tabsSequence) - 1
        while seqIndex >= 0:
            self.__tabsSequence[seqIndex] = \
                self.__tabsSequence[seqIndex] - stripCount
            seqIndex -= 1

        # Strip some items in the history
        self.__history = self.__history[-1 * self.limit:]

    def updateForCurrentIndex(self):
        """Called when the current tab is left"""
        if self.__index < 0:
            return

        # The file could be saved under the different name and get even a
        # different type, so update everything
        uuid = self.__history[self.__index].uuid
        widget = self.__editorsManger.getWidgetByUUID(uuid)

        self.__history[self.__index].tabType = widget.getType()
        self.__history[self.__index].displayName = widget.getShortName()

        if self.__history[self.__index].tabType in \
                            [MainWindowTabWidgetBase.PlainTextEditor,
                             MainWindowTabWidgetBase.VCSAnnotateViewer]:
            self.__history[self.__index].line = widget.getLine()
            self.__history[self.__index].pos = widget.getPos()
            self.__history[self.__index].firstVisible = \
                                    widget.getEditor().firstVisibleLine()
        else:
            self.__history[self.__index].line = -1
            self.__history[self.__index].pos = -1
            self.__history[self.__index].firstVisible = -1

        tabIndex = self.__editorsManger.getIndexByUUID(uuid)
        self.__history[self.__index].icon = \
                                self.__editorsManger.tabIcon(tabIndex)

    def testAdjacent(self, index):
        """tests if an adjacent history item is the same as the given"""
        if index > 0:
            if self.__history[index] == self.__history[index - 1]:
                return True
        if index < len(self.__history) - 1:
            if self.__history[index] == self.__history[index + 1]:
                return True
        return False

    def testAdjacentSeq(self, index):
        """tests if an adjacent seq item is the same as the given"""
        if index > 0:
            if self.__tabsSequence[index] == self.__tabsSequence[index - 1]:
                return True
        if index < len(self.__tabsSequence) - 1:
            if self.__tabsSequence[index] == self.__tabsSequence[index + 1]:
                return True
        return False

    def tabClosed(self, uuid):
        """Called when a tab is closed"""
        # Test if the current was closed
        oldCurrentIndex = self.__index

        # Build a list of history entries which were closed
        removedIndexes = []
        index = len(self.__history) - 1
        while index >= 0:
            if self.__history[index].uuid == uuid or \
               self.testAdjacent(index):
                removedIndexes.insert(0, index)
                del self.__history[index]
                if index < self.__index:
                    self.__index -= 1
            index -= 1

        if len(self.__history) == 0:
            # No history any more
            self.clear()
            return

        # Remove all such entries from the tabs seq and adjust indexes of those
        # which survived
        seqIndex = len(self.__tabsSequence) - 1
        while seqIndex >= 0:
            if self.__tabsSequence[seqIndex] in removedIndexes:
                del self.__tabsSequence[seqIndex]
            else:
                oldIndex = self.__tabsSequence[seqIndex]
                for item in removedIndexes:
                    if item < oldIndex:
                        self.__tabsSequence[seqIndex] -= 1
            seqIndex -= 1

        # Remove adjacent same entries in the tabs sequence
        seqIndex = len(self.__tabsSequence) - 1
        while seqIndex >= 0:
            if self.testAdjacentSeq(seqIndex):
                del self.__tabsSequence[seqIndex]
            seqIndex -= 1

        if oldCurrentIndex in removedIndexes:
            # Update it to the last visible
            self.__index = self.__tabsSequence[len(self.__tabsSequence) - 1]

        self.historyChanged.emit()

    def getSize(self):
        """Provides the total number of the history steps"""
        return len(self.__history)

    def getCurrentIndex(self):
        """Provides the current history index"""
        return self.__index

    def getEntry(self, index):
        """Provides the required history entry"""
        if index < 0 or index >= len(self.__history):
            raise Exception("Invalid history index to set (" +
                            str(index) + ")")
        return self.__history[index]

    def setCurrentIndex(self, index):
        """Sets the given history index as current"""
        if index < 0 or index >= len(self.__history):
            raise Exception("Invalid history index to set (" +
                            str(index) + ")")
        if self.__index != index:
            self.__index = index
            self.__tabsSequence.append(index)
            self.historyChanged.emit()

    def getCurrentEntry(self):
        """Provides the current history entry"""
        if self.__index == -1 or self.__index >= len(self.__history):
            raise Exception("No current history entry (index=" +
                            str(self.__index) + ")")
        return self.__history[self.__index]

    def stepBack(self):
        """Makes one step back in the history if possible"""
        if self.__index <= 0:
            return False
        self.__index -= 1
        self.__tabsSequence.append(self.__index)
        self.historyChanged.emit()
        return True

    def stepForward(self):
        """Makes one step forward in the history if possible"""
        if self.__index >= len(self.__history) - 1:
            return False
        self.__index += 1
        self.__tabsSequence.append(self.__index)
        self.historyChanged.emit()
        return True

    def flip(self):
        """Flips between last two history steps"""
        if len(self.__history) < 1:
            return False
        lastSeqIndex = len(self.__tabsSequence) - 1
        self.__index = self.__tabsSequence[lastSeqIndex - 1]
        self.__tabsSequence.append(self.__index)
        self.historyChanged.emit()
        return True

    def updateFileNameForTab(self, uuid, newFileName):
        """After SaveAs the file name should be updated"""
        newDisplayName = os.path.basename(newFileName)
        for item in self.__history:
            if item.uuid == uuid:
                item.displayName = newDisplayName

    def updateIconForTab(self, uuid, icon):
        """Broken/disappeared/modified icons could appear for a file"""
        for item in self.__history:
            if item.uuid == uuid:
                item.icon = icon



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy sergey.satskiy@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Custom calltips"""


from utils.globals import GlobalData
from .qt import (Qt, QEventLoop, QSizePolicy, QFrame, QLabel, QApplication,
                 QGridLayout)


class Calltip(QFrame):

    """Frameless panel with a calltip"""

    def __init__(self, parent):
        QFrame.__init__(self, parent)

        # Make the frame nice looking
        palette = self.palette()
        palette.setColor(self.backgroundRole(),
                         GlobalData().skin['calltipPaper'])
        self.setPalette(palette)

        self.setFrameShape(QFrame.StyledPanel)
        self.setLineWidth(2)
        self.setAutoFillBackground(True)

        # Keep pylint happy
        self.__calltipLabel = None
        self.__text = None
        self.__paramPositions = None
        self.__highlightedParam = None

        self.__createLayout()
        QFrame.hide(self)
        self.setFocusPolicy(Qt.NoFocus)

    def __createLayout(self):
        """Creates the widget layout"""
        self.__calltipLabel = QLabel('')
        self.__calltipLabel.setSizePolicy(QSizePolicy.Ignored,
                                          QSizePolicy.Fixed)
        self.__calltipLabel.setWordWrap(False)
        self.__calltipLabel.setAlignment(Qt.AlignLeft)
        palette = self.__calltipLabel.palette()
        palette.setColor(self.foregroundRole(),
                         GlobalData().skin['calltipColor'])
        self.__calltipLabel.setPalette(palette)

        gridLayout = QGridLayout(self)
        gridLayout.setContentsMargins(3, 3, 3, 3)
        gridLayout.addWidget(self.__calltipLabel, 0, 0, 1, 1)

        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

    def resize(self):
        """Resizes the dialogue to match the editor size"""
        vscroll = self.parent().verticalScrollBar()
        if vscroll.isVisible():
            scrollWidth = vscroll.width()
        else:
            scrollWidth = 0

        hscroll = self.parent().horizontalScrollBar()
        if hscroll.isVisible():
            scrollHeight = hscroll.height()
        else:
            scrollHeight = 0

        # Panel
        width = self.parent().width()
        height = self.parent().height()
        widgetWidth = width - scrollWidth - 6 - 1

        self.setFixedWidth(widgetWidth)

        vPos = height - self.height() - scrollHeight
        self.move(3, vPos - 3)

    def showCalltip(self, message, paramNumber):
        """Brings up the panel with the required text"""
        self.__text = message
        self.__calcParamPositions()
        self.highlightParameter(paramNumber)
        QApplication.processEvents(QEventLoop.ExcludeUserInputEvents)

        self.resize()
        self.show()

        # I have to do resize() twice because the self.height() value is
        # calculated in a different way before the frame is shown and after.
        # So the first call moves the widget to the nearly right location
        # and the second to the precise one
        self.resize()

    def highlightParameter(self, number):
        """Hightlights the given parameter number, 0 - based"""
        if number == self.__highlightedParam:
            return
        if self.__paramPositions is None:
            self.__calltipLabel.setText(self.__text)
            return
        if number >= len(self.__paramPositions):
            self.__calltipLabel.setText(self.__text)
            return
        positions = self.__paramPositions[number]
        highlight = self.__text[0:positions[0]] + "<font color='" + \
                    GlobalData().skin['calltipHighColor'].name() + "'>" + \
                    self.__text[positions[0]:positions[1] + 1] + \
                    "</font>" + self.__text[positions[1] + 1:]
        self.__calltipLabel.setText(highlight)

    def __calcParamPositions(self):
        """Calculates the parameter positions in the calltip text"""
        if self.__text is None or '\n' in self.__text:
            self.__paramPositions = None
            return

        try:
            begin = self.__text.index('(') + 1
        except:
            self.__paramPositions = None
            return

        if self.__text[begin] == '.':
            # Special case for f(...)
            self.__paramPositions = None
            return

        self.__paramPositions = []
        lastIndex = len(self.__text) - 1
        index = begin
        level = 0               # Unconditional skip of commas
        singleQuote = False
        doubleQuote = False
        while index <= lastIndex:
            ch = self.__text[index]
            if ch == "'" and singleQuote:
                singleQuote = False
            elif ch == '"' and doubleQuote:
                doubleQuote = False
            elif ch == "'":
                singleQuote = False
            elif ch == '"':
                doubleQuote = False
            elif ch in ['(', '{']:
                level += 1
            elif ch in [')', '}']:
                level -= 1
                if level == -1:
                    # Closing bracket
                    if index > begin:
                        self.__paramPositions.append((begin, index - 1))
                    break
            elif ch == '[':
                if level > 0:
                    index += 1
                    continue
                # if a previous char is '=' then it is a default paam
                checkIndex = index - 1
                while self.__text[checkIndex].isspace():
                    checkIndex -= 1
                if self.__text[checkIndex] == '=':
                    level += 1
                    index += 1
                    continue

                if self.__text[checkIndex] == '(':
                    # The very first round bracket
                    index = index + 1
                    begin = index
                    continue

                # '[' after a parameter name
                if index > begin:
                    self.__paramPositions.append((begin, index - 1))
                    # The next meaningfull character is comma or an identifier
                    index += 1
                    while index <= lastIndex and self.__text[index].isspace():
                        index += 1
                    if self.__text[index] == ',':
                        index += 1
                    index += 1
                    begin = index
                    continue

                # I don't know what it is
                self.__paramPositions = None
                return

            elif ch == ']':
                if level > 0:
                    level -= 1
                    index += 1
                    continue

                # This must be an optional argument closing bracket
                checkIndex = index - 1
                while self.__text[checkIndex].isspace():
                    checkIndex -= 1

                if self.__text[checkIndex] == ']':
                    index += 1
                    continue

                # Need to add and this is the last parameter
                if index > begin:
                    self.__paramPositions.append((begin, index - 1))
                break

            elif ch == ',':
                if level == 0:
                    self.__paramPositions.append((begin, index - 1))
                # Skip till the beginning of the next
                # parameter - it must be there
                index += 1
                while index <= lastIndex and self.__text[index].isspace():
                    index += 1
                begin = index
                continue

            index += 1

        if len(self.__paramPositions) == 0:
            self.__paramPositions = None

    def hide(self):
        """Handles the hiding of the panel and markers"""
        QFrame.hide(self)
        self.__text = None
        self.__paramPositions = None
        self.__highlightedParam = None
        self.__calltipLabel.setText("")



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Diff viewer tab widget"""

from utils.settings import Settings
from .qt import Qt, QEvent, QApplication
from .mainwindowtabwidgetbase import MainWindowTabWidgetBase
from .htmltabwidget import HTMLTabWidget


class DiffTabWidget(HTMLTabWidget):

    """The widget which displays a RO diff page"""

    def __init__(self, parent=None):
        HTMLTabWidget.__init__(self, parent)
        self.installEventFilter(self)

    def eventFilter(self, obj, event):
        """Event filter to catch shortcuts on UBUNTU"""
        if event.type() == QEvent.KeyPress:
            key = event.key()
            modifiers = event.modifiers()
            if modifiers == Qt.ControlModifier:
                if key == Qt.Key_Minus:
                    Settings().onZoomOut()
                    return True
                if key == Qt.Key_Equal:
                    Settings().onZoomIn()
                    return True
                if key == Qt.Key_0:
                    Settings().onZoomReset()
                    return True
            if modifiers == Qt.KeypadModifier | Qt.ControlModifier:
                if key == Qt.Key_Minus:
                    Settings().onZoomOut()
                    return True
                if key == Qt.Key_Plus:
                    Settings().onZoomIn()
                    return True
                if key == Qt.Key_0:
                    Settings().onZoomReset()
                    return True
        return HTMLTabWidget.eventFilter(self, obj, event)

    def wheelEvent(self, event):
        """Mouse wheel event"""
        if QApplication.keyboardModifiers() == Qt.ControlModifier:
            angleDelta = event.angleDelta()
            if not angleDelta.isNull():
                if angleDelta.y() > 0:
                    Settings().onZoomIn()
                else:
                    Settings().onZoomOut()
            event.accept()
        else:
            HTMLTabWidget.wheelEvent(self, event)

    def setHTML(self, content):
        """Sets the content from the given string"""
        HTMLTabWidget.setHTML(self, content)
        self.onTextZoomChanged()

    def loadFormFile(self, path):
        """Loads the content from the given file"""
        HTMLTabWidget.loadFormFile(self, path)
        self.zoomTo(Settings().zoom)

    def getType(self):
        """Tells the widget type"""
        return MainWindowTabWidgetBase.DiffViewer

    def getLanguage(self):
        """Tells the content language"""
        return "diff"



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2015-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Auxiliary items on a canvas which do not derive from CellElement"""

# pylint: disable=C0305
# pylint: disable=R0902
# pylint: disable=R0913


from sys import maxsize
from ui.qt import (QPen, QBrush, QPainterPath, Qt, QGraphicsSvgItem,
                   QGraphicsSimpleTextItem, QGraphicsRectItem,
                   QGraphicsPathItem, QGraphicsLineItem)
from .cellelement import CellElement


class SpacerCell(CellElement):

    """A cell which may take some space horizontally or vertically"""

    def __init__(self, ref, canvas, x, y, width=None, height=None):
        CellElement.__init__(self, ref, canvas, x, y)
        self.kind = CellElement.SPACER
        self.width = width
        self.height = height

    def render(self):
        """Renders the cell"""
        if self.width is None:
            self.width = self.canvas.settings.hSpacer
        if self.height is None:
            self.height = self.canvas.settings.vSpacer
        self.minWidth = self.width
        self.minHeight = self.height
        return (self.width, self.height)

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY


class VacantCell(SpacerCell):

    """A vacant cell which can be later used for some other element"""

    def __init__(self, ref, canvas, x, y):
        SpacerCell.__init__(self, ref, canvas, x, y, width=0, height=0)
        self.kind = CellElement.VACANT


class VSpacerCell(SpacerCell):

    """Represents a vertical spacer cell"""

    def __init__(self, ref, canvas, x, y, height=None):
        SpacerCell.__init__(
            self, ref, canvas, x, y, width=0,
            height=canvas.settings.vSpacer if height is None else height)
        self.kind = CellElement.V_SPACER


class HSpacerCell(SpacerCell):

    """Represents a horizontal spacer cell"""

    def __init__(self, ref, canvas, x, y, width=None):
        SpacerCell.__init__(
            self, ref, canvas, x, y,
            width=canvas.settings.hSpacer if width is None else width,
            height=0)
        self.kind = CellElement.H_SPACER


class SVGItem(CellElement, QGraphicsSvgItem):

    """Wrapper for an SVG items on the control flow"""

    def __init__(self, canvas, fName, ref):
        CellElement.__init__(self, ref, canvas, x=None, y=None)
        self.kind = CellElement.SVG

        self.__fName = fName
        QGraphicsSvgItem.__init__(self, self.__getPath(fName))
        self.__scale = 0

    @staticmethod
    def __getPath(fName):
        """Tries to resolve the given file name"""
        from utils.pixmapcache import getPixmapLocation
        path = getPixmapLocation(fName)
        if path is not None:
            return path
        return ''

    def setIconHeight(self, height):
        """Scales the svg item to the required height"""
        rectHeight = float(self.boundingRect().height())
        if rectHeight != 0.0:
            self.__scale = float(height) / rectHeight
            self.setScale(self.__scale)

    def setIconWidth(self, width):
        """Scales the svg item to the required width"""
        rectWidth = float(self.boundingRect().width())
        if rectWidth != 0.0:
            self.__scale = float(width) / rectWidth
            self.setScale(self.__scale)

    def iconHeight(self):
        """Provides the height"""
        return self.boundingRect().height() * self.__scale

    def iconWidth(self):
        """Provides the width"""
        return self.boundingRect().width() * self.__scale

    def mouseDoubleClickEvent(self, event):
        """Handling double click events"""
        if self.ref.kind == CellElement.EXCEPT_MINIMIZED:
            return self.ref.mouseDoubleClickEvent(event)
        return QGraphicsSvgItem.mouseDoubleClickEvent(self, event)

    def getSelectTooltip(self):
        """Provides the tooltip"""
        return "SVG picture for " + self.__fName

    def getProxiedItem(self):
        """Provides the real item for the proxy one"""
        return self.ref


class BadgeItem(CellElement, QGraphicsRectItem):

    """Serves the scope badges"""

    def __init__(self, ref, text):
        CellElement.__init__(self, ref, ref.canvas, None, None)
        QGraphicsRectItem.__init__(self)
        self.kind = CellElement.BADGE
        self.__text = text

        s = ref.canvas.settings

        self.__textRect = s.badgeFontMetrics.boundingRect(0, 0, maxsize,
                                                          maxsize, 0, text)

        self.width = self.__textRect.width() + 2 * s.badgeHSpacing
        self.height = self.__textRect.height() + 2 * s.badgeVSpacing

        self.__bgColor = s.badgeBGColor
        self.__fgColor = s.badgeFGColor
        self.__frameColor = s.badgeBorderColor
        self.__needRect = True

    def setBGColor(self, bgColor):
        """Sets the background color"""
        self.__bgColor = bgColor

    def setFGColor(self, fgColor):
        """Sets the foreground color"""
        self.__fgColor = fgColor

    def setFrameColor(self, frameColor):
        """Sets the frame color"""
        self.__frameColor = frameColor

    def setNeedRectangle(self, value):
        """Sets the need rectangle flag"""
        self.__needRect = value

    def moveTo(self, xPos, yPos):
        """Moves to the specified position"""
        # This is a mistery. I do not understand why I need to divide by 2.0
        # however this works. I tried various combinations of initialization,
        # setting the position and mapping. Nothing works but ../2.0. Sick!
        self.setPos(float(xPos) / 2.0, float(yPos) / 2.0)
        self.setRect(float(xPos) / 2.0, float(yPos) / 2.0,
                     self.width, self.height)

    def withinHeader(self):
        """True if it is within a header"""
        if self.ref.kind in [self.ref.ELSE_SCOPE,
                             self.ref.FINALLY_SCOPE,
                             self.ref.TRY_SCOPE]:
            return True
        if self.ref.kind == self.ref.EXCEPT_SCOPE:
            return self.ref.ref.clause is None
        return False

    def paint(self, painter, option, widget):
        """Paints the badge item"""
        del option
        del widget

        s = self.ref.canvas.settings

        if self.__needRect:
            pen = QPen(self.__frameColor)
            pen.setWidth(s.badgeLineWidth)
            painter.setPen(pen)
            painter.setBrush(QBrush(self.__bgColor))
            painter.drawRoundedRect(self.x(), self.y(),
                                    self.width, self.height,
                                    s.badgeRadius, s.badgeRadius)

        pen = QPen(self.__fgColor)
        painter.setPen(pen)
        painter.setFont(s.badgeFont)
        painter.drawText(self.x() + s.badgeHSpacing,
                         self.y() + s.badgeVSpacing,
                         self.__textRect.width(),
                         self.__textRect.height(),
                         Qt.AlignCenter, self.__text)

    def getSelectTooltip(self):
        """Provides the tooltip"""
        return "Badge '" + self.__text + "'"

    def getProxiedItem(self):
        """Provides the real item for a proxy one"""
        return self.ref



class Connector(CellElement, QGraphicsPathItem):

    """Implementation of a connector item"""

    def __init__(self, canvas, x1, y1, x2, y2):
        CellElement.__init__(self, None, canvas, x=None, y=None)
        QGraphicsPathItem.__init__(self)
        self.kind = CellElement.DEPENDENT_CONNECTOR

        path = QPainterPath()
        path.moveTo(x1, y1)
        path.lineTo(x2, y2)
        self.setPath(path)

        self.penStyle = None
        self.penColor = None
        self.penWidth = None

    def paint(self, painter, option, widget):
        """Paints the connector"""
        color = self.canvas.settings.cfLineColor
        if self.penColor:
            color = self.penColor
        width = self.canvas.settings.cfLineWidth
        if self.penWidth:
            width = self.penWidth

        pen = QPen(color)
        pen.setWidth(width)
        pen.setCapStyle(Qt.FlatCap)
        pen.setJoinStyle(Qt.RoundJoin)
        if self.penStyle:
            pen.setStyle(self.penStyle)
        self.setPen(pen)
        QGraphicsPathItem.paint(self, painter, option, widget)

    def getLastPoint(self):
        """Provides the last point"""
        path = self.path()
        lastElement = path.elementAt(path.elementCount() - 1)
        return (lastElement.x, lastElement.y)

    def getFirstPoint(self):
        """Provides the last point"""
        path = self.path()
        firstElement = path.elementAt(0)
        return (firstElement.x, firstElement.y)

    @staticmethod
    def getSelectTooltip():
        """Provides the tooltip"""
        return 'Connector'



class RubberBandItem(CellElement, QGraphicsRectItem):

    """Custom rubber band for selection"""

    def __init__(self, canvas):
        CellElement.__init__(self, None, canvas, x=None, y=None)
        QGraphicsRectItem.__init__(self)

        self.kind = CellElement.RUBBER_BAND
        self.__settings = canvas.settings
        self.__x = None
        self.__y = None
        self.__width = None
        self.__height = None

    def setGeometry(self, rect):
        """Sets the geometry"""
        # This is a mistery. I do not understand why I need to divide by 2.0
        # however this works. I tried various combinations of initialisation,
        # setting the position and mapping. Nothing works but ../2.0. Sick!
        self.__x = rect.x() / 2.0
        self.__y = rect.y() / 2.0
        self.__width = rect.width()
        self.__height = rect.height()

        self.setPos(self.__x, self.__y)
        self.setRect(self.__x, self.__y, self.__width, self.__height)

        self.update()

    def paint(self, painter, option, widget):
        """Paints the rubber band"""
        del option
        del widget

        pen = QPen(self.__settings.rubberBandBorderColor)
        painter.setPen(pen)
        painter.setBrush(QBrush(self.__settings.rubberBandFGColor))
        painter.drawRect(self.__x, self.__y,
                         self.__width, self.__height)



class Text(CellElement, QGraphicsSimpleTextItem):

    """Implementation of a text item"""

    def __init__(self, canvas, text):
        CellElement.__init__(self, None, canvas, None, None)
        QGraphicsSimpleTextItem.__init__(self)
        self.kind = CellElement.TEXT

        self.setFont(canvas.settings.badgeFont)
        self.setText(text)

        self.color = None

    def paint(self, painter, option, widget):
        """Paints the text item"""
        color = self.canvas.settings.cfLineColor
        if self.color:
            color = self.color

        self.setBrush(QBrush(color))
        QGraphicsSimpleTextItem.paint(self, painter, option, widget)

    @staticmethod
    def getSelectTooltip():
        """Provides the tooltip"""
        return 'Text'



class Line(CellElement, QGraphicsLineItem):

    """Implementation of the line item"""

    def __init__(self, canvas, x1, y1, x2, y2):
        CellElement.__init__(self, None, canvas, x=None, y=None)
        QGraphicsLineItem.__init__(self, x1, y1, x2, y2)
        self.kind = CellElement.LINE

        self.penStyle = None
        self.penColor = None
        self.penWidth = None

    def paint(self, painter, option, widget):
        """Paints the line item"""
        color = self.canvas.settings.cfLineColor
        if self.penColor:
            color = self.penColor
        width = self.canvas.settings.cfLineWidth
        if self.penWidth:
            width = self.penWidth

        pen = QPen(color)
        pen.setWidth(width)
        pen.setCapStyle(Qt.FlatCap)
        pen.setJoinStyle(Qt.RoundJoin)
        if self.penStyle:
            pen.setStyle(self.penStyle)
        self.setPen(pen)

        QGraphicsLineItem.paint(self, painter, option, widget)

    @staticmethod
    def getSelectTooltip():
        """Provides the tooltip"""
        return 'Line'


class Rectangle(CellElement, QGraphicsRectItem):

    """Implementation of the rectangle item"""

    def __init__(self, canvas, x, y, w, h):
        CellElement.__init__(self, None, canvas, x=None, y=None)
        QGraphicsRectItem.__init__(self, x, y, w, h)
        self.kind = CellElement.RECTANGLE

        self.pen = None
        self.brush = None

    def paint(self, painter, option, widget):
        """Paints the rectangle item"""
        if self.pen is not None:
            self.setPen(self.pen)
        if self.brush is not None:
            self.setBrush(self.brush)

        QGraphicsRectItem.paint(self, painter, option, widget)

    @staticmethod
    def getSelectTooltip():
        """Provides the tooltip"""
        return "Rectangle"



class ConnectorCell(CellElement, QGraphicsPathItem):

    """Represents a single connector cell"""

    NORTH = 0
    SOUTH = 1
    WEST = 2
    EAST = 3
    CENTER = 4

    # Connector type. In case of 'if' and groups it is necessary to calculate
    # properly how wide the connector should be. The subKind tells what kind
    # of correction is required
    GENERIC = 100
    TOP_IF = 101
    BOTTOM_IF = 102

    def __init__(self, connections, canvas, x, y):
        """Connections are supposed to be a list of tuples.

        Eample: [ (NORTH, SOUTH), (EAST, CENTER) ]
        """
        CellElement.__init__(self, None, canvas, x, y)
        QGraphicsPathItem.__init__(self)
        self.kind = CellElement.CONNECTOR
        self.subKind = self.GENERIC
        self.connections = connections

    def hasVertical(self):
        """True if has a vertical part"""
        for conn in self.connections:
            if self.NORTH in conn or self.SOUTH in conn:
                return True
        return False

    def hasHorizontal(self):
        """True if has a horizontal part"""
        for conn in self.connections:
            if self.EAST in conn or self.WEST in conn:
                return True
        return False

    def render(self):
        """Renders the cell"""
        settings = self.canvas.settings

        if self.hasVertical():
            self.minWidth = settings.mainLine + settings.hCellPadding
            self.minWidth += self.hShift * 2 * settings.openGroupHSpacer
        else:
            self.minWidth = 0

        if self.hasHorizontal():
            self.minHeight = 2 * settings.vCellPadding
        else:
            self.minHeight = 0

        self.height = self.minHeight
        self.width = self.minWidth
        return (self.width, self.height)

    def __getY(self):
        """Provides the Y coordinate"""
        row = self.addr[1]
        column = self.addr[0]
        cells = self.canvas.cells
        for index in range(column - 1, -1, -1):
            cell = cells[row][index]
            if cell.isSpacerItem():
                continue
            if cell.scopedItem():
                break
            if cell.kind != CellElement.CONNECTOR:
                return cell.minHeight / 2
        return self.height / 2

    def __getNorthXY(self, baseX):
        """Provides the north coordinates"""
        settings = self.canvas.settings
        if self.subKind == self.BOTTOM_IF:
            cellAbove = self.canvas.cells[self.addr[1] - 1][self.addr[0]]
            if cellAbove.kind == CellElement.VCANVAS:
                baseX += cellAbove.maxGlobalOpenGroupDepth * 2 * \
                         settings.openGroupHSpacer
        return baseX + settings.mainLine, self.baseY

    def __getSouthXY(self, baseX):
        """Provides the south coordinates"""
        settings = self.canvas.settings
        if self.subKind == self.TOP_IF:
            cellBelow = self.canvas.cells[self.addr[1] + 1][self.addr[0]]
            if cellBelow.kind == CellElement.VCANVAS:
                baseX += cellBelow.maxGlobalOpenGroupDepth * 2 * \
                         settings.openGroupHSpacer
        return baseX + settings.mainLine, self.baseY + self.height

    def __getXY(self, location):
        """Provides the location coordinates"""
        settings = self.canvas.settings
        hShift = self.hShift * 2 * settings.openGroupHSpacer

        baseX = self.baseX
        if self.subKind not in [self.TOP_IF, self.BOTTOM_IF]:
            baseX = self.baseX + hShift

        if location == self.NORTH:
            return self.__getNorthXY(baseX)
        if location == self.SOUTH:
            return self.__getSouthXY(baseX)
        if location == self.WEST:
            return baseX, self.baseY + self.__getY()
        if location == self.EAST:
            return baseX + self.width - hShift, self.baseY + self.__getY()

        # It is CENTER
        if self.subKind == self.TOP_IF:
            cellBelow = self.canvas.cells[self.addr[1] + 1][self.addr[0]]
            if cellBelow.kind == CellElement.VCANVAS:
                baseX += cellBelow.maxGlobalOpenGroupDepth * 2 * \
                         settings.openGroupHSpacer
        elif self.subKind == self.BOTTOM_IF:
            cellAbove = self.canvas.cells[self.addr[1] - 1][self.addr[0]]
            if cellAbove.kind == CellElement.VCANVAS:
                baseX += cellAbove.maxGlobalOpenGroupDepth * 2 * \
                         settings.openGroupHSpacer
        return baseX + settings.mainLine, self.baseY + self.__getY()

    def __angled(self, begin, end):
        """Returns True if the connection is not straight"""
        if begin in [self.NORTH, self.SOUTH] and \
           end in [self.WEST, self.EAST]:
            return True
        return end in [self.NORTH, self.SOUTH] and \
               begin in [self.WEST, self.EAST]

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY

        path = QPainterPath()
        for connection in self.connections:
            startX, startY = self.__getXY(connection[0])
            endX, endY = self.__getXY(connection[1])
            if self.__angled(connection[0], connection[1]):
                centerX, centerY = self.__getXY(self.CENTER)
                path.moveTo(startX, startY)
                path.lineTo(centerX, centerY)
                path.lineTo(endX, endY)
            else:
                path.moveTo(startX, startY)
                path.lineTo(endX, endY)
        # It does not look nice so commented out
        #if len(self.connections) == 1:
        #    if self.connections[0][0] == self.NORTH:
        #        if self.connections[0][1] == self.CENTER:
        #            # That's a half connector used when terminal items are
        #            # suppressed.
        #            radius = self.canvas.settings.vCellPadding / 2.0
        #            path.addEllipse(endX - radius / 2.0,
        #                            endY - radius / 2.0, radius, radius)
        self.setPath(path)
        scene.addItem(self)

    def paint(self, painter, option, widget):
        """Draws the code block"""
        settings = self.canvas.settings

        pen = QPen(settings.cfLineColor)
        pen.setWidth(settings.cfLineWidth)
        pen.setJoinStyle(Qt.RoundJoin)
        self.setPen(pen)
        painter.setPen(pen)
        QGraphicsPathItem.paint(self, painter, option, widget)




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Update interval settings dialog"""

from ui.qt import (Qt, QDialog, QVBoxLayout, QDialogButtonBox, QHBoxLayout,
                   QLabel, QLineEdit, QIntValidator)


class VCSUpdateIntervalConfigDialog(QDialog):

    """Dialog to configure update interval"""

    def __init__(self, value, parent=None):
        QDialog.__init__(self, parent)
        self.interval = value

        self.__createLayout()
        self.setWindowTitle("VCS file status update interval configuration")

        self.__intervalEdit.setText(str(self.interval))
        self.__updateOKStatus()

        self.__intervalEdit.textChanged.connect(self.__updateOKStatus)
        self.__intervalEdit.setFocus()

    def __createLayout(self):
        """Creates the dialog layout"""
        self.resize(400, 80)
        self.setSizeGripEnabled(True)

        vboxLayout = QVBoxLayout(self)

        hboxLayout = QHBoxLayout()
        hboxLayout.addWidget(QLabel("Status update interval, sec."))
        self.__intervalEdit = QLineEdit()
        self.__intervalEdit.setValidator(QIntValidator(1, 3600, self))
        self.__intervalEdit.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        hboxLayout.addWidget(self.__intervalEdit)

        # Buttons at the bottom
        self.__buttonBox = QDialogButtonBox(self)
        self.__buttonBox.setOrientation(Qt.Horizontal)
        self.__buttonBox.setStandardButtons(QDialogButtonBox.Ok |
                                            QDialogButtonBox.Cancel)
        self.__buttonBox.accepted.connect(self.userAccept)
        self.__buttonBox.rejected.connect(self.close)

        vboxLayout.addLayout(hboxLayout)
        vboxLayout.addWidget(self.__buttonBox)

    def __updateOKStatus(self):
        """Updates the OK button status"""
        okButton = self.__buttonBox.button(QDialogButtonBox.Ok)

        if self.__intervalEdit.text() == "":
            okButton.setEnabled(False)
            okButton.setToolTip("Interval must be defined")
            return

        value = int(self.__intervalEdit.text())
        if value < 1 or value > 3600:
            okButton.setEnabled(False)
            okButton.setToolTip("Interval must be within 1..3600 sec")
            return

        okButton.setEnabled(True)
        okButton.setToolTip("")

    def userAccept(self):
        """Triggered when the user clicks OK"""
        self.interval = int(self.__intervalEdit.text())
        self.accept()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Various editor buffer related utilities"""

import re
import time
from cdmpyparser import getBriefModuleInfoFromMemory


WORD_PATTERN = '\w+'
WORD_REGEXP = re.compile(WORD_PATTERN)
EDITOR_TAG_TIMEOUT = 0.5


def getEditorTags(editor, exclude=None, excludePythonKeywords=False):
    """Builds a list of the tags in the editor.

    The current line could be excluded.
    The only tags are included which start with prefix
    """
    start = time.time()

    excludeSet = set()
    if exclude:
        excludeSet.add(exclude)
    if excludePythonKeywords:
        # Note: 2 characters words will be filtered unconditionally
        excludeSet.update(["try", "for", "and", "not"])

    result = set()
    for line in editor.lines:
        for match in WORD_REGEXP.findall(line):
            if len(match) > 2:
                if match not in excludeSet:
                    result.add(match)
        if time.time() - start > EDITOR_TAG_TIMEOUT:
            # Do not freeze longer than that
            break

    # If a cursor is in a middle of the word then the current word is not what
    # you need.
    if editor.getWordAfterCursor() == '':
        result.discard(editor.getWordBeforeCursor())
    return result


class TextCursorContext:

    """Holds the text cursor context for a python file"""

    GlobalScope = 1
    FunctionScope = 2
    ClassScope = 3
    ClassMethodScope = 4

    def __init__(self):
        self.levels = []    # Each item is [infoObj, scope type]
        self.length = 0

    def addFunction(self, infoObj):
        """Adds nested function"""
        if self.length == 0:
            self.levels.append([infoObj, self.FunctionScope])
        else:
            if self.levels[self.length - 1][1] == self.ClassScope:
                self.levels.append([infoObj, self.ClassMethodScope])
            else:
                self.levels.append([infoObj, self.FunctionScope])
        self.length += 1

    def getScope(self):
        """Provides the deepest scope type"""
        if self.length == 0:
            return self.GlobalScope
        return self.levels[self.length - 1][1]

    def getInfoObj(self):
        """Provides the deepest info object"""
        if self.length == 0:
            return None
        return self.levels[self.length - 1][0]

    def addClass(self, infoObj):
        """Adds nested class"""
        self.levels.append([infoObj, self.ClassScope])
        self.length += 1

    def __scopeToString(self, scope):
        """Converts scope constant to a string"""
        if scope == self.GlobalScope:
            return "GlobalScope"
        if scope == self.FunctionScope:
            return "FunctionScope"
        if scope == self.ClassScope:
            return "ClassScope"
        if scope == self.ClassMethodScope:
            return "ClassMethodScope"
        return "UnknownScope"

    def __str__(self):
        """Converts context to a string representation"""
        retval = ""
        if self.length == 0:
            retval = "GlobalScope"

        first = True
        for level in self.levels:
            if first:
                first = False
            else:
                retval += " -> "
            retval += self.__scopeToString(level[1]) + \
                ":" + level[0].name + ":" + str(level[0].line)
        return retval

    def getLastScopeLine(self):
        """Provides the last scope line"""
        if self.length == 0:
            raise Exception("No scopes found")
        return self.levels[self.length - 1][0].colonLine

    def stripLevels(self, nonSpacePos):
        """Strips the levels depending on the position"""
        maxLevels = int(nonSpacePos / 4)
        if maxLevels < self.length:
            self.levels = self.levels[:maxLevels]
            self.length = maxLevels


def _IdentifyScope(infoObject, context, cursorLine, cursorPos, skipDef):
    """Searches for the hierarchy"""
    # Find the closest class definition (global level for the first call)
    nearestClassLine = -1
    nearestClassInfo = None
    for klass in infoObject.classes:
        onDef = _isOnDefinitionLine(klass, cursorLine, cursorPos)
        if skipDef:
            if onDef:
                return
            if klass.line > nearestClassLine and klass.line < cursorLine:
                nearestClassLine = klass.line
                nearestClassInfo = klass
        else:
            if onDef:
                context.addClass(klass)
                return
            if klass.line > nearestClassLine and klass.line < cursorLine:
                nearestClassLine = klass.line
                nearestClassInfo = klass

    # Find the closest function definition (global level for the first call)
    nearestFuncLine = -1
    nearestFuncInfo = None
    for func in infoObject.functions:
        onDef = _isOnDefinitionLine(func, cursorLine, cursorPos)
        if skipDef:
            if onDef:
                return
            if func.line > nearestClassLine and \
               func.line > nearestFuncLine and \
               func.line <= cursorLine:
                nearestFuncLine = func.line
                nearestFuncInfo = func
        else:
            if onDef:
                context.addFunction(func)
                return
            if func.line > nearestClassLine and \
               func.line > nearestFuncLine and \
               func.line <= cursorLine:
                nearestFuncLine = func.line
                nearestFuncInfo = func

    if nearestClassLine == -1 and nearestFuncLine == -1:
        # No definitions before the line
        return

    # Check nested objects
    if nearestClassLine > nearestFuncLine:
        context.addClass(nearestClassInfo)
        _IdentifyScope(nearestClassInfo, context,
                       cursorLine, cursorPos, skipDef)
    else:
        context.addFunction(nearestFuncInfo)
        _IdentifyScope(nearestFuncInfo, context,
                       cursorLine, cursorPos, skipDef)


def _getFirstNonSpacePos(text):
    """Provides the index of the first non-space character in the given line"""
    for pos in range(len(text)):
        if text[pos] not in [' ', '\n', '\r']:
            return pos
    return -1


def getContext(editor, info=None,
               skipBlankLinesBack=False, skipDef=True):
    """Detects the context at the text cursor position.

    skipBlankLinesBack == False => current cursor position is used
    skipBlankLinesBack == True => skip blank lines back and use the first
                                  non-blank line as the cursor position.
    skipDef == True => treat a definition as belonging to an upper
                       level context (not included into the context stack)
    skipDef == False => treat a definition as starting a context level
                        (included into the context stack as the last one)
    """
    # It is expected that this is a python editor.
    # If non-python editor is given, then a global context is provided

    context = TextCursorContext()

    if not editor.isPythonBuffer():
        return context

    # It's not the first position, so the parsed module info is required
    if info is None:
        info = getBriefModuleInfoFromMemory(editor.text)

    line, pos = editor.cursorPosition
    if skipBlankLinesBack:
        while line >= 0:
            text = editor.lines[line]
            trimmedText = text.strip()
            if trimmedText != "":
                pos = len(text.rstrip())
                break
            line -= 1
        if line < 0:
            line = 0
            pos = 0

    _IdentifyScope(info, context, line + 1, pos, skipDef)

    if not skipDef:
        if _getDefinitionObject(info, line + 1, pos) is not None:
            return context

    if context.length == 0:
        return context

    continueLine = False
    currentLine = context.getLastScopeLine() + 1
    for currentLine in range(context.getLastScopeLine(),
                             len(editor.lines)):
        if currentLine == line:
            break

        text = editor.lines[currentLine]
        textLen = len(text)
        trimmedText = text.strip()
        if not continueLine:
            if trimmedText == "" or trimmedText.startswith("#"):
                continue

            # Here: there must be characters in the line
            nonSpacePos = _getFirstNonSpacePos(text)
            context.stripLevels(nonSpacePos)
            if context.length == 0:
                return context

        if trimmedText.endswith(",") or trimmedText.endswith('\\') or \
           (textLen > 0 and editor.isStringLiteral(currentLine, textLen - 1)):
            continueLine = True
        else:
            continueLine = False

    if continueLine:
        context.stripLevels(nonSpacePos)
    else:
        nonSpacePos = _getFirstNonSpacePos(editor.lines[line])
        if nonSpacePos == -1:
            context.stripLevels(pos)
        else:
            context.stripLevels(min(pos, nonSpacePos))
    return context


def _isOnDefinitionLine(infoObj, line, pos):
    """True if the cursor is within the definition line of the infoObj.

    infoObj is a class or function
    Line and pos are 1-based
    """
    lowLimit = infoObj.keywordLine << 16
    upLimit = (infoObj.colonLine << 16) + infoObj.colonPos
    current = (line << 16) + pos
    for decor in infoObj.decorators:
        candidate = (decor.line << 16) + decor.pos
        if candidate < lowLimit:
            lowLimit = candidate
    if current >= lowLimit and current <= upLimit:
        return True
    return False


def _getDefinitionObject(info, line, pos):
    """Returns a class or a function if the cursor is on the definition line.

    Line and pos are 1-based
    """
    for cls in info.classes:
        if _isOnDefinitionLine(cls, line, pos):
            return cls
        obj = _getDefinitionObject(cls, line, pos)
        if obj:
            return obj
    for func in info.functions:
        if _isOnDefinitionLine(func, line, pos):
            return func
        obj = _getDefinitionObject(func, line, pos)
        if obj:
            return obj
    return None


def getItemForDisplayPath(info, displayPath):
    """Info is what the parser provides.

    displayPath is a list of what displayed in a tree.
    The method provides the certain item from the info if it is still there
    """
    # Ugly but helps to avoid initialization obstacles
    from ui.viewitems import (FunctionItemType, ClassesItemType,
                              FunctionsItemType, ImportsItemType,
                              InstanceAttributesItemType,
                              StaticAttributesItemType, GlobalsItemType,
                              CodingItemType, ImportWhatItemType,
                              DecoratorItemType)
    for (itemType, pathItem) in displayPath:
        if itemType == ClassesItemType:
            info = info.classes
        elif itemType == FunctionsItemType:
            info = info.functions
        elif itemType == ImportsItemType:
            info = info.imports
        elif itemType == InstanceAttributesItemType:
            info = info.instanceAttributes
        elif itemType == StaticAttributesItemType:
            info = info.classAttributes
        elif itemType == GlobalsItemType:
            info = info.globals
        elif itemType == CodingItemType:
            return info.encoding
        else:
            # That's a name, find it in the container
            if itemType == ImportWhatItemType:
                info = info.what
            elif itemType == FunctionItemType:
                if not isinstance(info, list):
                    info = info.functions
            elif itemType == DecoratorItemType:
                info = info.decorators
            found = False
            for item in info:
                if item.getDisplayName() == pathItem:
                    found = True
                    info = item
                    break
            if found:
                continue
            return None
    return info



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Watch points viewer"""

from ui.qt import (Qt, pyqtSignal, QModelIndex, QSizePolicy, QFrame, QTreeView,
                   QToolButton, QHeaderView, QVBoxLayout, QItemSelectionModel,
                   QLabel, QWidget, QAbstractItemView, QMenu, QSpacerItem,
                   QHBoxLayout, QPalette, QSortFilterProxyModel)
from ui.itemdelegates import NoOutlineHeightDelegate
from utils.pixmapcache import getIcon
from utils.globals import GlobalData
from utils.project import CodimensionProject
from utils.settings import Settings


class WatchPointView(QTreeView):

    """Watch expression viewer widget"""

    sigSelectionChanged = pyqtSignal(QModelIndex)

    def __init__(self, parent, wpointsModel):
        QTreeView.__init__(self, parent)

        self.__model = None
        self.setModel(wpointsModel)

        self.setItemsExpandable(False)
        self.setRootIsDecorated(False)
        self.setAlternatingRowColors(True)
        self.setUniformRowHeights(True)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setItemDelegate(NoOutlineHeightDelegate(4))

        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.__showContextMenu)
        self.doubleClicked.connect(self.__doubleClicked)

        self.__createPopupMenus()

    def setModel(self, model):
        """Sets the watch expression model"""
        self.__model = model

        self.sortingModel = QSortFilterProxyModel()
        self.sortingModel.setSourceModel(self.__model)
        QTreeView.setModel(self, self.sortingModel)

        header = self.header()
        header.setSortIndicator(0, Qt.AscendingOrder)
        header.setSortIndicatorShown(True)
        header.setSectionsClickable(True)

        self.setSortingEnabled(True)
        self.__layoutDisplay()

    def __layoutDisplay(self):
        """Performs the layout operation"""
        self.__resizeColumns()
        self.__resort()

    def __resizeColumns(self):
        """Resizes the view when items get added, edited or deleted"""
        self.header().resizeSections(QHeaderView.ResizeToContents)
        self.header().setStretchLastSection(True)

    def __resort(self):
        """Resorts the tree"""
        self.model().sort(self.header().sortIndicatorSection(),
                          self.header().sortIndicatorOrder())

    def __toSourceIndex(self, index):
        """Converts an index to a source index"""
        return self.sortingModel.mapToSource(index)

    def __fromSourceIndex(self, sindex):
        """Converts a source index to an index"""
        return self.sortingModel.mapFromSource(sindex)

    def __setRowSelected(self, index, selected=True):
        """Selects a row"""
        if not index.isValid():
            return

        if selected:
            flags = QItemSelectionModel.SelectionFlags(
                QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows)
        else:
            flags = QItemSelectionModel.SelectionFlags(
                QItemSelectionModel.Deselect | QItemSelectionModel.Rows)
        self.selectionModel().select(index, flags)

    def __createPopupMenus(self):
        """Generates the popup menus"""
        self.menu = QMenu()
        self.menu.addAction("Add", self.__addWatchPoint)
        self.menu.addAction("Edit...", self.__editWatchPoint)
        self.menu.addSeparator()
        self.menu.addAction("Enable", self.__enableWatchPoint)
        self.menu.addAction("Enable all", self.__enableAllWatchPoints)
        self.menu.addSeparator()
        self.menu.addAction("Disable", self.__disableWatchPoint)
        self.menu.addAction("Disable all", self.__disableAllWatchPoints)
        self.menu.addSeparator()
        self.menu.addAction("Delete", self.__deleteWatchPoint)
        self.menu.addAction("Delete all", self.__deleteAllWatchPoints)

        self.backMenuActions = {}
        self.backMenu = QMenu()
        self.backMenu.addAction("Add", self.__addWatchPoint)
        self.backMenuActions["EnableAll"] = \
            self.backMenu.addAction("Enable all", self.__enableAllWatchPoints)
        self.backMenuActions["DisableAll"] = \
            self.backMenu.addAction("Disable all",
                                    self.__disableAllWatchPoints)
        self.backMenuActions["DeleteAll"] = \
            self.backMenu.addAction("Delete all",
                                    self.__deleteAllWatchPoints)
        self.backMenu.aboutToShow.connect(self.__showBackMenu)

        self.multiMenu = QMenu()
        self.multiMenu.addAction("Add", self.__addWatchPoint)
        self.multiMenu.addSeparator()
        self.multiMenu.addAction("Enable selected",
                                 self.__enableSelectedWatchPoints)
        self.multiMenu.addAction("Enable all", self.__enableAllWatchPoints)
        self.multiMenu.addSeparator()
        self.multiMenu.addAction("Disable selected",
                                 self.__disableSelectedWatchPoints)
        self.multiMenu.addAction("Disable all", self.__disableAllWatchPoints)
        self.multiMenu.addSeparator()
        self.multiMenu.addAction("Delete selected",
                                 self.__deleteSelectedWatchPoints)
        self.multiMenu.addAction("Delete all", self.__deleteAllWatchPoints)

    def __showContextMenu(self, coord):
        """Show the context menu"""
        cnt = self.__getSelectedItemsCount()
        if cnt <= 1:
            index = self.indexAt(coord)
            if index.isValid():
                cnt = 1
                self.__setRowSelected(index)
        coord = self.mapToGlobal(coord)
        if cnt > 1:
            self.multiMenu.popup(coord)
        elif cnt == 1:
            self.menu.popup(coord)
        else:
            self.backMenu.popup(coord)

    def __findDuplicates(self, cond, special,
                         showMessage=False, index=QModelIndex()):
        """Checks if an entry already exists"""
        idx = self.__model.getWatchPointIndex(cond, special)
        duplicate = idx.isValid() and idx.internalPointer() != index.internalPointer()
#        if showMessage and duplicate:
#            if not special:
#                msg = """<p>A watch expression '<b>%1</b>'"""
#                                  """ already exists.</p>""".arg(Utilities.html_encode(unicode(cond)))
#            else:
#                msg = self.trUtf8("""<p>A watch expression '<b>%1</b>'"""
#                                  """ for the variable <b>%2</b> already exists.</p>""")\
#                        .arg(special)\
#                        .arg(Utilities.html_encode(unicode(cond)))
#            KQMessageBox.warning(None,
#                "Watch expression already exists", msg)
        return duplicate

    def __clearSelection(self):
        """Clears the selection"""
        for index in self.selectedIndexes():
            self.__setRowSelected(index, False)

    def __addWatchPoint(self):
        """Adds watch expression via a context menu entry"""
#        dlg = EditWatchpointDialog( ( "", False, True, 0, "" ), self )
#        if dlg.exec_() == QDialog.Accepted:
#            cond, temp, enabled, ignorecount, special = dlg.getData()
#            if not self.__findDuplicates(cond, special, True):
#                self.__model.addWatchPoint(cond, special, (temp, enabled, ignorecount))
#                self.__resizeColumns()
#                self.__resort()
        return

    def __doubleClicked(self, index):
        """Handles the double clicked signal"""
        if index.isValid():
            self.__doEditWatchPoint(index)

    def __editWatchPoint(self):
        """Handles the edit watch expression context menu entry"""
        index = self.currentIndex()
        if index.isValid():
            self.__doEditWatchPoint(index)

    def __doEditWatchPoint(self, index):
        """Edits a watch expression"""
        sindex = self.__toSourceIndex(index)
        if sindex.isValid():
            wp = self.__model.getWatchPointByIndex(sindex)
            if not wp:
                return

            cond, special, temp, enabled, count = wp[:5]

#            dlg = EditWatchpointDialog(
#                (cond, temp, enabled, count, special), self)
#            if dlg.exec_() == QDialog.Accepted:
#                cond, temp, enabled, count, special = dlg.getData()
#                if not self.__findDuplicates(cond, special, True, sindex):
#                    self.__model.setWatchPointByIndex(sindex,
#                        unicode(cond), unicode(special), (temp, enabled, count))
#                    self.__resizeColumns()
#                    self.__resort()
        return

    def __setWpEnabled(self, index, enabled):
        """Sets the enabled status of a watch expression"""
        sindex = self.__toSourceIndex(index)
        if sindex.isValid():
            self.__model.setWatchPointEnabledByIndex(sindex, enabled)

    def __enableWatchPoint(self):
        """Handles the enable watch expression context menu entry"""
        index = self.currentIndex()
        self.__setWpEnabled(index, True)
        self.__resizeColumns()
        self.__resort()

    def __enableAllWatchPoints(self):
        """Handles the enable all watch expressions context menu entry"""
        index = self.model().index(0, 0)
        while index.isValid():
            self.__setWpEnabled(index, True)
            index = self.indexBelow(index)
        self.__resizeColumns()
        self.__resort()

    def __enableSelectedWatchPoints(self):
        """Handles the enable selected watch expressions context menu entry"""
        for index in self.selectedIndexes():
            if index.column() == 0:
                self.__setWpEnabled(index, True)
        self.__resizeColumns()
        self.__resort()

    def __disableWatchPoint(self):
        """Handles the disable watch expression context menu entry"""
        index = self.currentIndex()
        self.__setWpEnabled(index, False)
        self.__resizeColumns()
        self.__resort()

    def __disableAllWatchPoints(self):
        """Handles the disable all watch expressions context menu entry"""
        index = self.model().index(0, 0)
        while index.isValid():
            self.__setWpEnabled(index, False)
            index = self.indexBelow(index)
        self.__resizeColumns()
        self.__resort()

    def __disableSelectedWatchPoints(self):
        """Handles the disable selected watch expressions context menu entry"""
        for index in self.selectedIndexes():
            if index.column() == 0:
                self.__setWpEnabled(index, False)
        self.__resizeColumns()
        self.__resort()

    def __deleteWatchPoint(self):
        """Handles the delete watch expression context menu entry"""
        index = self.currentIndex()
        sindex = self.__toSourceIndex(index)
        if sindex.isValid():
            self.__model.deleteWatchPointByIndex(sindex)

    def __deleteAllWatchPoints(self):
        """Handles the delete all watch expressions context menu entry"""
        self.__model.deleteAll()

    def __deleteSelectedWatchPoints(self):
        """Handles the delete selected watch expressions context menu entry"""
        idxList = []
        for index in self.selectedIndexes():
            sindex = self.__toSourceIndex(index)
            if sindex.isValid() and index.column() == 0:
                lastrow = index.row()
                idxList.append(sindex)
        self.__model.deleteWatchPoints(idxList)

    def __showBackMenu(self):
        """Handles the aboutToShow signal of the background menu"""
        if self.model().rowCount() == 0:
            self.backMenuActions["EnableAll"].setEnabled(False)
            self.backMenuActions["DisableAll"].setEnabled(False)
            self.backMenuActions["DeleteAll"].setEnabled(False)
        else:
            self.backMenuActions["EnableAll"].setEnabled(True)
            self.backMenuActions["DisableAll"].setEnabled(True)
            self.backMenuActions["DeleteAll"].setEnabled(True)

    def __getSelectedItemsCount(self):
        """Provides the count of items selected"""
        count = len(self.selectedIndexes()) / (self.__model.columnCount() - 1)
        # column count is 1 greater than selectable
        return count

    def selectionChanged(self, selected, deselected):
        """The slot is called when the selection has changed"""
        if selected.indexes():
            self.sigSelectionChanged.emit(selected.indexes()[0])
        else:
            self.sigSelectionChanged.emit(QModelIndex())
        QTreeView.selectionChanged(self, selected, deselected)


class WatchPointViewer(QWidget):

    """Implements the watch point viewer for a debugger"""

    def __init__(self, parent, wpointModel):
        QWidget.__init__(self, parent)

        self.__currentItem = None

        self.__createPopupMenu()
        self.__createLayout(wpointModel)

        GlobalData().project.sigProjectChanged.connect(self.__onProjectChanged)

        if Settings()['showWatchPointViewer'] == False:
            self.__onShowHide(True)

    def __createPopupMenu(self):
        """Creates the popup menu"""
#        self.__excptMenu = QMenu()
#        self.__removeMenuItem = self.__excptMenu.addAction(
#                    "Remove from ignore list", self.__onRemoveFromIgnore )
        return

    def __createLayout(self, wpointModel):
        """Creates the widget layout"""
        verticalLayout = QVBoxLayout(self)
        verticalLayout.setContentsMargins(0, 0, 0, 0)
        verticalLayout.setSpacing(0)

        self.headerFrame = QFrame()
        self.headerFrame.setFrameStyle(QFrame.StyledPanel)
        self.headerFrame.setAutoFillBackground(True)
        headerPalette = self.headerFrame.palette()
        headerBackground = headerPalette.color(QPalette.Background)
        headerBackground.setRgb(min(headerBackground.red() + 30, 255),
                                min(headerBackground.green() + 30, 255),
                                min(headerBackground.blue() + 30, 255))
        headerPalette.setColor(QPalette.Background, headerBackground)
        self.headerFrame.setPalette(headerPalette)
        self.headerFrame.setFixedHeight(24)

        self.__watchpointLabel = QLabel("Watchpoints")

        expandingSpacer = QSpacerItem(10, 10, QSizePolicy.Expanding)
        fixedSpacer = QSpacerItem(3, 3)

        self.__showHideButton = QToolButton()
        self.__showHideButton.setAutoRaise(True)
        self.__showHideButton.setIcon(getIcon('less.png'))
        self.__showHideButton.setFixedSize(20, 20)
        self.__showHideButton.setToolTip("Hide ignored exceptions list")
        self.__showHideButton.setFocusPolicy(Qt.NoFocus)
        self.__showHideButton.clicked.connect(self.__onShowHide)

        headerLayout = QHBoxLayout()
        headerLayout.setContentsMargins(1, 1, 1, 1)
        headerLayout.addSpacerItem(fixedSpacer)
        headerLayout.addWidget(self.__watchpointLabel)
        headerLayout.addSpacerItem(expandingSpacer)
        headerLayout.addWidget(self.__showHideButton)
        self.headerFrame.setLayout(headerLayout)

        self.__wpointsList = WatchPointView(self, wpointModel)

        self.__enableButton = QToolButton()
        self.__enableButton.setIcon(getIcon('add.png'))
        self.__enableButton.setFixedSize(24, 24)
        self.__enableButton.setToolTip("Enable/disable the watchpoint")
        self.__enableButton.setFocusPolicy(Qt.NoFocus)
        self.__enableButton.setEnabled(False)
        self.__enableButton.clicked.connect(self.__onEnableDisable)

        expandingSpacer = QSpacerItem(10, 10, QSizePolicy.Expanding)

        self.__jumpToCodeButton = QToolButton()
        self.__jumpToCodeButton.setIcon(getIcon('gotoline.png'))
        self.__jumpToCodeButton.setFixedSize(24, 24)
        self.__jumpToCodeButton.setToolTip("Jump to the code")
        self.__jumpToCodeButton.setFocusPolicy(Qt.NoFocus)
        self.__jumpToCodeButton.setEnabled(False)
        self.__jumpToCodeButton.clicked.connect(self.__onJumpToCode)

        toolbarLayout = QHBoxLayout()
        toolbarLayout.addWidget(self.__enableButton)
        toolbarLayout.addSpacerItem(expandingSpacer)
        toolbarLayout.addWidget(self.__jumpToCodeButton)

        self.__wpointsList.sigSelectionChanged.connect(
            self.__onSelectionChanged)

        verticalLayout.addWidget(self.headerFrame)
        verticalLayout.addLayout(toolbarLayout)
        verticalLayout.addWidget(self.__wpointsList)

    def clear(self):
        """Clears the content"""
#        self.__wpointsList.clear()
        self.__updateTitle()
        self.__jumpToCodeButton.setEnabled(False)
        self.__currentItem = None

    def __onJumpToCode(self):
        """Jumps to the corresponding source code line"""
        return

    def __onShowHide(self, startup=False):
        """Triggered when show/hide button is clicked"""
        if startup or self.__wpointsList.isVisible():
            self.__wpointsList.setVisible(False)
            self.__enableButton.setVisible(False)
            self.__jumpToCodeButton.setVisible(False)
            self.__showHideButton.setIcon(getIcon('more.png'))
            self.__showHideButton.setToolTip("Show watchpoints list")

            self.__minH = self.minimumHeight()
            self.__maxH = self.maximumHeight()

            self.setMinimumHeight(self.headerFrame.height())
            self.setMaximumHeight(self.headerFrame.height())

            Settings()['showWatchPointViewer'] = False
        else:
            self.__wpointsList.setVisible(True)
            self.__enableButton.setVisible(True)
            self.__jumpToCodeButton.setVisible(True)
            self.__showHideButton.setIcon(getIcon('less.png'))
            self.__showHideButton.setToolTip("Hide watchpoints list")

            self.setMinimumHeight(self.__minH)
            self.setMaximumHeight(self.__maxH)

            Settings()['showWatchPointViewer'] = True

    def __onSelectionChanged(self, index):
        """Triggered when the current item is changed"""
        if index.isValid():
            pass
        else:
            pass
        return
        selected = list(self.__exceptionsList.selectedItems())
        if selected:
            self.__currentItem = selected[0]
            self.__removeButton.setEnabled(True)
        else:
            self.__currentItem = None
            self.__removeButton.setEnabled(False)

    def __updateTitle(self):
        """Updates the section title"""
        count = self.getTotalCount()
        if count == 0:
            self.__watchpointLabel.setText("Watchpoints")
        else:
            self.__watchpointLabel.setText("Watchpoints (total: " +
                                           str(count) + ")")

    def getTotalCount(self):
        """Provides the total number of watch points"""
        count = 0
        return count

    def __onProjectChanged(self, what):
        """Triggered when a project is changed"""
        if what == CodimensionProject.CompleteProject:
            self.clear()

    def __onEnableDisable(self):
        """Triggered when a breakpoint should be enabled/disabled"""
        pass



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2015-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""break and continue cells"""

# pylint: disable=C0305
# pylint: disable=R0902
# pylint: disable=R0913

from ui.qt import QGraphicsRectItem, QPen, Qt, QGraphicsItem, QBrush
from .cellelement import CellElement
from .colormixin import ColorMixin
from .textmixin import TextMixin
from .auxitems import Connector


class LoopJumpBase(CellElement, TextMixin, ColorMixin, QGraphicsRectItem):

    """Base class for 'break' and 'continue'"""

    def __init__(self, ref, canvas, x, y, bgColor, fgColor, borderColor):
        CellElement.__init__(self, ref, canvas, x, y)
        TextMixin.__init__(self)
        ColorMixin.__init__(self, ref, bgColor, fgColor, borderColor)
        QGraphicsRectItem.__init__(self)

        self.connector = None

        # Cache for the size
        self.xPos = None
        self.yPos = None
        self.rectWidth = None
        self.rectHeight = None

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def renderCell(self, customText, hPadding, vPadding):
        """Renders the cell"""
        settings = self.canvas.settings
        self.setupText(self, customText=customText)

        vPadding = 2 * (vPadding + settings.vCellPadding)
        self.minHeight = self.textRect.height() + vPadding

        hPadding = 2 * (hPadding + settings.hCellPadding)
        self.minWidth = self.textRect.width() + hPadding

        if settings.noContent:
            self.minWidth = max(self.minWidth, settings.minWidth)
        self.height = self.minHeight
        self.width = self.minWidth
        return (self.width, self.height)

    def __calculateSize(self):
        """Calculates the size"""
        settings = self.canvas.settings
        self.xPos = self.baseX + settings.hCellPadding
        self.yPos = self.baseY + settings.vCellPadding
        self.rectWidth = self.minWidth - 2 * settings.hCellPadding
        self.rectHeight = self.minHeight - 2 * settings.hCellPadding

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY

        # Add the connector as a separate scene item to make the selection
        # working properly
        settings = self.canvas.settings
        self.connector = Connector(self.canvas,
                                   baseX + settings.mainLine, baseY,
                                   baseX + settings.mainLine,
                                   baseY + settings.vCellPadding)
        scene.addItem(self.connector)

        # Setting the rectangle is important for the selection and for
        # redrawing. Thus the selection pen with must be considered too.
        penWidth = settings.selectPenWidth - 1
        self.__calculateSize()
        self.setRect(self.xPos - penWidth, self.yPos - penWidth,
                     self.rectWidth + 2 * penWidth,
                     self.rectHeight + 2 * penWidth)
        scene.addItem(self)

    def paintCell(self, painter, rectRadius):
        """Draws the cell"""
        settings = self.canvas.settings
        painter.setPen(self.getPainterPen(self.isSelected(), self.borderColor))
        painter.setBrush(QBrush(self.bgColor))

        painter.drawRoundedRect(self.xPos, self.yPos,
                                self.rectWidth, self.rectHeight,
                                rectRadius, rectRadius)

        # Draw the text in the rectangle
        pen = QPen(self.fgColor)
        painter.setFont(settings.monoFont)
        painter.setPen(pen)

        hShift = (self.rectWidth - self.textRect.width()) / 2
        vShift = (self.rectHeight - self.textRect.height()) / 2
        painter.drawText(self.xPos + hShift, self.yPos + vShift,
                         self.textRect.width(), self.textRect.height(),
                         Qt.AlignLeft, self.text)


class BreakCell(LoopJumpBase):

    """Represents a single break statement"""

    def __init__(self, ref, canvas, x, y):
        LoopJumpBase.__init__(self, ref, canvas, x, y,
                              canvas.settings.breakBGColor,
                              canvas.settings.breakFGColor,
                              canvas.settings.breakBorderColor)
        self.kind = CellElement.BREAK

    def render(self):
        """Renders the cell"""
        return self.renderCell('break',
                               self.canvas.settings.breakHPadding,
                               self.canvas.settings.breakVPadding)

    def paint(self, painter, option, widget):
        """Draws the cell"""
        del option
        del widget

        self.paintCell(painter, self.canvas.settings.breakRectRadius)

    def getSelectTooltip(self):
        """Provides the tooltip"""
        return 'Break at ' + CellElement.getLinesSuffix(self.getLineRange())


class ContinueCell(LoopJumpBase):

    """Represents a single continue statement"""

    def __init__(self, ref, canvas, x, y):
        LoopJumpBase.__init__(self, ref, canvas, x, y,
                              canvas.settings.continueBGColor,
                              canvas.settings.continueFGColor,
                              canvas.settings.continueBorderColor)
        self.kind = CellElement.CONTINUE

    def render(self):
        """Renders the cell"""
        return self.renderCell('continue',
                               self.canvas.settings.continueHPadding,
                               self.canvas.settings.continueVPadding)

    def paint(self, painter, option, widget):
        """Draws the break statement"""
        del option      # unused argument
        del widget      # unused argument

        self.paintCell(painter, self.canvas.settings.continueRectRadius)

    def getSelectTooltip(self):
        """Provides the tooltip"""
        return "Continue at " + CellElement.getLinesSuffix(self.getLineRange())




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The ideas and code samples are taken from the winpdb project.
# Credits: Nir Aides, Copyright (C) 2005-2009
#

"""Utility functions to support running scripts"""

import sys
import os.path
from subprocess import check_output, STDOUT
from shlex import quote
from .config import DEFAULT_ENCODING
from .runparams import RUN, PROFILE, DEBUG
from .encoding import detectFileEncodingToRead


def prepareArguments(arguments):
    """Prepares arguments for the command line"""
    args = []
    for index, _ in enumerate(arguments):
        args.append('${CDM_ARG' + str(index) + '}')
    return args


def prepareInterpreter(params):
    """Provides the intrpreter path, quoted if needed"""
    if params['useInherited']:
        return quote(sys.executable)
    return quote(params['customInterpreter'])


def prepareScript(path):
    """Provides the scropt path quoted if needed"""
    return quote(path)


def getTerminalCommandToRun(fileName, arguments, params,
                            tcpServerPort=None, procuuid=None):
    """Provides a command to run a separate shell terminal"""
    interpreter = prepareInterpreter(params)
    script = prepareScript(fileName)

    if params['redirected']:
        runClient = os.path.sep.join([os.path.dirname(sys.argv[0]),
                                      "debugger", "client",
                                      "client_cdm_run.py"])
        parts = [interpreter, runClient,
                 '--host', 'localhost', '--port', str(tcpServerPort),
                 '--procuuid', str(procuuid), '--', script] + \
                prepareArguments(arguments)
        return ' '.join(parts)

    # Non-redirected case, i.e. the user provided the a custom terminal string
    parts = [interpreter, script] + prepareArguments(arguments)
    customTerminal = params['customTerminal'].strip()
    if '${prog}' in customTerminal:
        return customTerminal.replace('${prog}', ' '.join(parts))

    return customTerminal + ' ' + ' '.join(parts)


def getTerminalCommandToProfile(fileName, arguments, params,
                                tcpServerPort=None, procuuid=None):
    """Provides a command to run a separate shell terminal"""
    interpreter = prepareInterpreter(params)
    script = prepareScript(fileName)

    from .globals import GlobalData
    outfile = GlobalData().getProfileOutputPath(procuuid)

    if params['redirected']:
        runClient = os.path.sep.join([os.path.dirname(sys.argv[0]),
                                      "debugger", "client",
                                      "client_cdm_profile.py"])
        parts = [interpreter, runClient,
                 '--host', 'localhost', '--port', str(tcpServerPort),
                 '--procuuid', str(procuuid), '--outfile', outfile,
                 '--', script] + \
                prepareArguments(arguments)
        return ' '.join(parts)

    # Non-redirected case, i.e. the user provided the a custom terminal string
    parts = [interpreter, script] + prepareArguments(arguments)
    customTerminal = params['customTerminal'].strip()
    if '${prog}' in customTerminal:
        return customTerminal.replace('${prog}', ' '.join(parts))

    return customTerminal + ' ' + ' '.join(parts)


def getTerminalCommandToDebug(fileName, arguments, params,
                              tcpServerPort, procuuid):
    """Provides a command line to debug in a separate shell terminal"""
    interpreter = prepareInterpreter(params)
    script = prepareScript(fileName)
    encoding = detectFileEncodingToRead(fileName)

    debugClient = os.path.sep.join([os.path.dirname(sys.argv[0]),
                                    "debugger", "client",
                                    "client_cdm_dbg.py"])
    parts = [interpreter, debugClient,
             '--host', 'localhost', '--port', str(tcpServerPort),
             '--procuuid', str(procuuid), '--encoding', encoding]

    # Get the debugging specific parameters
    from .settings import Settings
    debugSettings = Settings().getDebuggerSettings()

    # Form the debug client options
    if not debugSettings.reportExceptions:
        parts.append('--no-exc-report')
    if debugSettings.traceInterpreter:
        parts.append('--trace-python')
    if debugSettings.autofork:
        if debugSettings.followChild:
            parts.append('--fork-child')
        else:
            parts.append('--fork-parent')

    if not Settings()['calltrace']:
        parts.append('--no-call-trace')

    if params['redirected']:
        parts += ['--', script] + prepareArguments(arguments)
        return ' '.join(parts)

    # Append the option of a non redirected I/O
    parts.append('--no-redirect')
    parts += ['--', script] + prepareArguments(arguments)

    customTerminal = params['customTerminal'].strip()
    if '${prog}' in customTerminal:
        return customTerminal.replace('${prog}', ' '.join(parts))

    return customTerminal + ' ' + ' '.join(parts)


def parseCommandLineArguments(cmdLine):
    """Parses command line arguments provided by the user in the UI"""
    result = []

    cmdLine = cmdLine.strip()
    expectQuote = False
    expectDblQuote = False
    lastIndex = len(cmdLine) - 1
    argument = ""
    index = 0
    while index <= lastIndex:
        if expectQuote:
            if cmdLine[index] == "'":
                if cmdLine[index - 1] != '\\':
                    if argument != "":
                        result.append(argument)
                        argument = ""
                    expectQuote = False
                else:
                    argument = argument[: -1] + "'"
            else:
                argument += cmdLine[index]
            index += 1
            continue
        if expectDblQuote:
            if cmdLine[index] == '"':
                if cmdLine[index - 1] != '\\':
                    if argument != "":
                        result.append(argument)
                        argument = ""
                    expectDblQuote = False
                else:
                    argument = argument[: -1] + '"'
            else:
                argument += cmdLine[index]
            index += 1
            continue
        # Not in a string literal
        if cmdLine[index] == "'":
            if index == 0 or cmdLine[index - 1] != '\\':
                expectQuote = True
                if argument != "":
                    result.append(argument)
                    argument = ""
            else:
                argument = argument[: -1] + "'"
            index += 1
            continue
        if cmdLine[index] == '"':
            if index == 0 or cmdLine[index - 1] != '\\':
                expectDblQuote = True
                if argument != "":
                    result.append(argument)
                    argument = ""
            else:
                argument = argument[: -1] + '"'
            index += 1
            continue
        if cmdLine[index] in (' ', '\t'):
            if argument != "":
                result.append(argument)
                argument = ""
            index += 1
            continue
        argument += cmdLine[index]
        index += 1

    if argument != "":
        result.append(argument)

    if expectQuote or expectDblQuote:
        raise Exception("No closing quotation")
    return result


def getCwdCmdEnv(kind, path, params, tcpServerPort=None, procuuid=None):
    """Provides the working directory, command line and environment.

    It covers all: running/debugging/profiling a script
    """
    # The arguments parsing is going to pass OK because it
    # was checked in the run parameters dialogue
    if kind not in [RUN, PROFILE, DEBUG]:
        raise Exception("Unknown command requested. "
                        "Supported command types are: run, profile, debug.")

    arguments = parseCommandLineArguments(params['arguments'])
    if kind == RUN:
        cmd = getTerminalCommandToRun(path, arguments, params,
                                      tcpServerPort, procuuid)
    elif kind == PROFILE:
        cmd = getTerminalCommandToProfile(path, arguments, params,
                                          tcpServerPort, procuuid)
    elif kind == DEBUG:
        cmd = getTerminalCommandToDebug(path, arguments, params,
                                        tcpServerPort, procuuid)

    environment = getNoArgsEnvironment(params)
    for index, value in enumerate(arguments):
        environment['CDM_ARG' + str(index)] = value

    return cmd, environment


def getNoArgsEnvironment(params):
    """Provides a copy of the environment"""
    if params['envType'] == params.InheritParentEnv:
        # 'None' does not work here: popen stores last env somewhere and
        # uses it inappropriately
        return os.environ.copy()
    if params['envType'] == params.InheritParentEnvPlus:
        environment = os.environ.copy()
        environment.update(params['additionToParentEnv'])
        return environment
    return params['specificEnv'].copy()


def checkOutput(cmdLine, useShell=False):
    """Wrapper around Subprocess.check_output which respects encoding"""
    if useShell:
        if not isinstance(cmdLine, str):
            raise Exception("Running via shell requires "
                            "the command line as a string")
    else:
        if not isinstance(cmdLine, list):
            raise Exception("Running without shell requires "
                            "the command line as a list")
    return check_output(cmdLine, stderr=STDOUT,
                        shell=useShell).decode(DEFAULT_ENCODING)


if __name__ == '__main__':
    print("Current working dir: " + os.getcwd())
    print("Environment: " + str(os.environ))
    print("Arguments: " + str(sys.argv))



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2019  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Text mixin for the graphics items"""

# pylint: disable=C0305

from html import escape
from .cml import CMLVersion, CMLrt
from .cellelement import CellElement


class TextMixin:

    """Text mixin to support text displayed by the item and a tooltip"""

    def __init__(self):
        self.text = None            # Text to display; it may be:
                                    # - text from the code
                                    # - replacement text if so
                                    # - nothing if suppressed
        self.textRect = None

    @staticmethod
    def getReplacementText(ref):
        """Provides the replacement text if so from the CML comment"""
        try:
            rtComment = CMLVersion.find(ref.leadingCMLComments, CMLrt)
            if rtComment:
                return rtComment.getText()
        except AttributeError:
            # The leadingCMLComments attribute may be not available
            return None
        return None

    @staticmethod
    def __getDisplayText(graphicsItem, customText):
        """Provides the display text"""
        if customText is None:
            return graphicsItem.ref.getDisplayValue()
        return customText

    @staticmethod
    def __shouldHideText(graphicsItem):
        """True if the text should be hidden"""
        settings = graphicsItem.canvas.settings
        if settings.hidecomments:
            if graphicsItem.isComment() or graphicsItem.isCMLDoc():
                return True

        if settings.noContent:
            if graphicsItem.kind not in [CellElement.CLASS_SCOPE,
                                         CellElement.FUNC_SCOPE]:
                return True

        return False

    def setupText(self, graphicsItem, customText=None, customReplacement=None):
        """Prepares the text and its rectangle; sets tooltip if needed"""
        if customReplacement is None:
            replacement = TextMixin.getReplacementText(graphicsItem.ref)
        else:
            replacement = customReplacement

        if TextMixin.__shouldHideText(graphicsItem):
            # The content needs to be suppressed
            tooltip = '<pre>' + \
                      TextMixin.__getDisplayText(graphicsItem, customText) + \
                      '</pre>'

            if replacement:
                tooltip += '<hr>Replacement text:<pre>' + \
                    escape(replacement) + '</pre>'
            graphicsItem.setToolTip(tooltip)

            self.text = ''
        else:
            # No suppression; something needs to be shown
            self.text = TextMixin.__getDisplayText(graphicsItem, customText)
            if replacement:
                graphicsItem.setToolTip('<pre>' + escape(self.text) + '</pre>')
                self.text = replacement
        self.textRect = graphicsItem.getBoundingRect(self.text)

